import os, sys, argparse
import pandas as pd
import numpy as np
import cwas.utils.log as log
from pathlib import Path
from tqdm import tqdm
#from glmnet import ElasticNet, LogitNet
from sklearn.linear_model import ElasticNetCV
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import r2_score
from cwas.core.common import cmp_two_arr
from cwas.utils.check import check_is_file, check_num_proc
from cwas.runnable import Runnable
from typing import Optional, Tuple
from contextlib import contextmanager
from collections import defaultdict
import matplotlib.pyplot as plt
import polars as pl
import re

class RiskScore(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._sample_info = None
        self._categorization_result = None
        self._adj_factor = None
        self._category_set_path = None
        self._category_set = None
        self._datasets = None
        self._covariates = None
        self._test_covariates = None
        self._response = None
        self._test_response = None
        self._result_dict = defaultdict(dict)
        self._permutation_dict = defaultdict(dict)
        self._filtered_combs = None
    
    @staticmethod
    def _print_args(args: argparse.Namespace):
        log.print_arg(
            "Categorization result file", 
            args.categorization_result_path
                if args.categorization_result_path
                else "Not specified: $CATEGORIZATION_RESULT will be used")
        log.print_arg("Sample information file", args.sample_info_path)
        log.print_arg("Adjustment factor list", args.adj_factor_path)
        log.print_arg(
            "Category set file", 
            args.category_set_path)
        if args.tag:
            log.print_arg("Output tag (prefix of output files)", args.tag)
        log.print_arg("If the number of carriers is used for calculating R2 or not", args.use_n_carrier)
        log.print_arg(
            "Threshold for selecting rare categories",
            f"{args.ctrl_thres: ,d}",
        )
        log.print_arg("Fraction of the training set", f"{args.train_set_f: ,f}")
        log.print_arg(
            "No. regression trials to calculate a mean of R squares",
            f"{args.num_reg: ,d}",
        )
        log.print_arg(
            "No. folds for CV",
            f"{args.fold: ,d}",
        )
        #log.print_arg("Use Logistic regression", args.logistic)
        log.print_arg(
            "No. permutation used to calculate the p-value",
            f"{args.n_permute: ,d}",
        )
        log.print_arg("Skip the permutation test", args.predict_only)
        log.print_arg(
            "No. worker processes for permutation",
            f"{args.num_proc: ,d}",
        )

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.sample_info_path)
        check_num_proc(args.num_proc)
        if args.categorization_result_path:
            check_is_file(args.categorization_result_path)
        if args.adj_factor_path is not None:
            check_is_file(args.adj_factor_path)
        if args.category_set_path:
            check_is_file(args.category_set_path)
    
    @property
    def categorization_result_path(self) -> Path:
        return (
            self.args.categorization_result_path.resolve()
            if self.args.categorization_result_path 
            else Path(self.get_env("CATEGORIZATION_RESULT"))
        )

    @property
    def adj_factor_path(self) -> Optional[Path]:
        return (
            self.args.adj_factor_path.resolve()
            if self.args.adj_factor_path
            else None
        )

    @property
    def plot_path(self) -> Optional[Path]:
        tag = '' if self.tag is None else ''.join([self.tag, '_'])
        f_name = re.sub(r'.categorization_result\.txt\.gz|.categorization_result\.txt', f'.lasso_histogram_{tag}thres_{self.ctrl_thres}.pdf', str(self.categorization_result_path.name))
        return Path(
            f"{self.out_dir}/" +
            f"{f_name}"
        )

    @property
    def category_set_path(self) -> Optional[Path]:
        return (
            self.args.category_set_path.resolve()
            if self.args.category_set_path
            else None
        )

    #@property
    #def logistic(self) -> bool:
    #    return self.args.logistic

    @property
    def out_dir(self) -> Path:
        return(self.args.output_dir_path.resolve())

    @property
    def category_set(self) -> pd.DataFrame:
        if self._category_set is None and self.category_set_path:
            self._category_set = pd.read_csv(self.category_set_path, sep='\t')
        return self._category_set

    @property
    def sample_info_path(self) -> Path:
        return self.args.sample_info_path.resolve()

    @property
    def use_n_carrier(self) -> bool:
        return self.args.use_n_carrier

    @property
    def domain_list(self) -> str:
        if self.args.domain_list=='run_all':
            return 'all,coding,noncoding,ptv,missense,damaging_missense,promoter,noncoding_wo_promoter,intron,intergenic,utr,lincRNA'
        else:
            return self.args.domain_list
        

    @property
    def tag(self) -> str:
        return self.args.tag

    @property
    def num_proc(self) -> int:
        return self.args.num_proc

    @property
    def num_reg(self) -> int:
        return self.args.num_reg

    @property
    def train_set_f(self) -> float:
        return self.args.train_set_f

    @property
    def fold(self) -> int:
        return self.args.fold

    @property
    def n_permute(self) -> int:
        return self.args.n_permute
    
    @property
    def ctrl_thres(self) -> int:
        return self.args.ctrl_thres
    
    @property
    def predict_only(self) -> bool:
        return self.args.predict_only
    
    @property
    def seed(self) -> int:
        return self.args.seed
    
    @property
    def sample_info(self) -> pd.DataFrame:
        if self._sample_info is None:
            self._sample_info = pd.read_table(
                self.sample_info_path, index_col="SAMPLE", dtype={"SAMPLE": str}, sep="\t"
            )
            if ("SET" not in self._sample_info.columns):
                log.print_log("LOG", 
                              "No 'SET' column in sample information file. "
                              "Training and test sets will be assigned randomly.")
                
                case_count = sum(self._sample_info['PHENOTYPE'] == 'case')
                ctrl_count = sum(self._sample_info['PHENOTYPE'] == 'ctrl')
                
                self.case_f = round(case_count*self.train_set_f)
                self.ctrl_f = round(ctrl_count*self.train_set_f)
                
                
                log.print_log("LOG",
                              "Use {} cases and {} controls for training set".format(self.case_f, self.ctrl_f)
                              )
                              #"Use {} samples from each phenotype as training set."
                              #.format(self.min_size))
                
                
                case_test_idx = self._sample_info.loc[self._sample_info.PHENOTYPE=='case'].sample(n=self.case_f, random_state=42).index
                ctrl_test_idx = self._sample_info.loc[self._sample_info.PHENOTYPE=='ctrl'].sample(n=self.ctrl_f, random_state=42).index
                self._sample_info.loc[case_test_idx, 'SET'] = 'training'
                self._sample_info.loc[ctrl_test_idx, 'SET'] = 'training'
                self._sample_info["SET"] = self._sample_info['SET'].fillna('test')
                
                #self.min_size = int(np.rint(min(case_count, ctrl_count) * self.train_set_f))
                #test_idx = self._sample_info.groupby('PHENOTYPE').sample(n=self.min_size, random_state=42).index
                #self._sample_info["SET"] = np.where(self._sample_info.index.isin(test_idx), "test", "training")
        return self._sample_info

    @property
    def adj_factor(self) -> pd.DataFrame:
        if self._adj_factor is None and self.adj_factor_path:
            self._adj_factor = pd.read_table(
                self.adj_factor_path, index_col="SAMPLE", dtype={"SAMPLE": str}, sep="\t"
            )
        return self._adj_factor
    
    @property
    def categorization_result(self) -> pd.DataFrame:
        if self._categorization_result is None:
            log.print_progress("Load the categorization result")
            self._categorization_result = pl.read_csv(
                self.categorization_result_path, dtypes={"SAMPLE": str}, separator = '\t'
            )
            self._categorization_result = self._categorization_result.to_pandas().set_index("SAMPLE")
            if self.adj_factor is not None:
                self._adjust_categorization_result()
            if self.use_n_carrier:
                self._categorization_result = self._categorization_result.applymap(lambda x: 1 if x > 0 else 0)
            log.print_log("LOG",
                          "Categorization result has {} samples and {} categories."
                          .format(self._categorization_result.shape[0], self._categorization_result.shape[1]))
        return self._categorization_result
    
    def _adjust_categorization_result(self):
        if not self._contain_same_index(
           self._categorization_result, self._adj_factor
        ):
            raise ValueError(
                "The sample IDs from the adjustment factor list are "
                "not the same with the sample IDs "
                "from the categorization result."
            )
        '''
        adj_factors = [
            self.adj_factor.to_dict()["AdjustFactor"][sample_id]
            for sample_id in self._categorization_result.index.values
        ]
        '''
        adj_factors = self.adj_factor.loc[self._categorization_result.index.values, "AdjustFactor"].tolist()
        self._categorization_result = self._categorization_result.multiply(
            adj_factors, axis="index"
        )

    @staticmethod
    def _contain_same_index(table1: pd.DataFrame, table2: pd.DataFrame) -> bool:
        return cmp_two_arr(table1.index.values, table2.index.values)
    
    @property
    def datasets(self) -> np.ndarray:
        if self._datasets is None:
            sample_ids = self.categorization_result.index.values
            set_dict = self.sample_info.to_dict()["SET"]
            self._datasets = np.array([set_dict[sample_id] for sample_id in sample_ids])
        return self._datasets

    @property
    def covariates(self) -> pd.DataFrame:
        if self._covariates is None:
            self._covariates = self.categorization_result[self.datasets == "training"]
        return self._covariates
    
    @property
    def test_covariates(self) -> pd.DataFrame:
        if self._test_covariates is None:
            self._test_covariates = self.categorization_result[self.datasets == "test"]
        return self._test_covariates
    
    @property
    def response(self) -> np.ndarray:
        if self._response is None:
            sample_ids = self.categorization_result[self.datasets == "training"].index.values
            phenotype_dict = self.sample_info.to_dict()["PHENOTYPE"]
            self._response = np.array([phenotype_dict[sample_id] == "case" for sample_id in sample_ids])
        return self._response

    @property
    def test_response(self) -> np.ndarray:
        if self._test_response is None:
            sample_ids = self.categorization_result[self.datasets == "test"].index.values
            phenotype_dict = self.sample_info.to_dict()["PHENOTYPE"]
            self._test_response = np.array([phenotype_dict[sample_id] == "case" for sample_id in sample_ids])
        return self._test_response

    @property
    def coef_path(self) -> Path:
        tag = '' if self.tag is None else ''.join([self.tag, '_'])
        f_name = re.sub(r'.categorization_result\.txt\.gz|.categorization_result\.txt', f'.lasso_coef_{tag}thres_{self.ctrl_thres}.txt', str(self.categorization_result_path.name))
        return Path(
            f"{self.out_dir}/" +
            f"{f_name}"
        )
    
    @property
    def result_path(self) -> Path:
        tag = '' if self.tag is None else ''.join([self.tag, '_'])
        f_name = re.sub(r'.categorization_result\.txt\.gz|.categorization_result\.txt', f'.lasso_results_{tag}thres_{self.ctrl_thres}.txt', str(self.categorization_result_path.name))
        return Path(
            f"{self.out_dir}/" +
            f"{f_name}"
        )

    @property
    def null_model_path(self) -> Path:
        tag = '' if self.tag is None else ''.join([self.tag, '_'])
        f_name = re.sub(r'.categorization_result\.txt\.gz|.categorization_result\.txt', f'.lasso_null_models_{tag}thres_{self.ctrl_thres}.txt', str(self.categorization_result_path.name))
        return Path(
            f"{self.out_dir}/" +
            f"{f_name}"
        )

    def run(self):
        self.prepare()
        self.risk_scores()
        if not self.predict_only:
            self.permute_pvalues()
        self.save_results()
        self.update_env()
        log.print_progress("Done")

    def prepare(self):
        if not self._contain_same_index(
            self.categorization_result, self.sample_info
        ):
            raise ValueError(
                "The sample IDs from the sample information are "
                "not the same with the sample IDs "
                "from the categorization result."
            )
            
    def risk_scores(self):
        """Generate risk scores for various seeds """
        log.print_progress(self.risk_scores.__doc__)

        domain_values = [domain.strip() for domain in self.domain_list.split(',')]
        for domain in domain_values:
            if domain == 'all':
                filtered_combs = pd.Series(self.categorization_result.columns)
            else:
                filtered_combs = self.category_set.loc[self.category_set['is_'+domain]==1]['Category']
            
            seeds = np.arange(self.seed, self.seed + self.num_reg * 10, 10)
            for seed in seeds:
                self.risk_score_per_category(domain = domain, result_dict=self._result_dict, seed=seed, filtered_combs=filtered_combs)
            
    def risk_score_per_category(self, domain: str, result_dict: defaultdict, seed: int = 42, swap_label: bool = False, filtered_combs = None):
        """Lasso model selection """
        if swap_label:
            np.random.seed(seed)
            
            perm_sample_info = self.sample_info
            perm_sample_info['Perm_PHENOTYPE'] = np.random.permutation(self.sample_info['PHENOTYPE'])
            perm_sample_info = perm_sample_info.drop(columns=['PHENOTYPE', 'SET'])
            
            perm_case_test_idx = perm_sample_info.loc[perm_sample_info.Perm_PHENOTYPE=='case'].sample(n=self.case_f, random_state=seed).index
            perm_ctrl_test_idx = perm_sample_info.loc[perm_sample_info.Perm_PHENOTYPE=='ctrl'].sample(n=self.ctrl_f, random_state=seed).index
            perm_sample_info.loc[perm_case_test_idx, 'Perm_SET'] = 'training'
            perm_sample_info.loc[perm_ctrl_test_idx, 'Perm_SET'] = 'training'
            perm_sample_info["Perm_SET"] = perm_sample_info['Perm_SET'].fillna('test')

            #test_idx = perm_sample_info.groupby('Perm_PHENOTYPE').sample(n=self.min_size, random_state=seed).index
            #perm_sample_info["Perm_SET"] = np.where(perm_sample_info.index.isin(test_idx), "test", "training")
            
            sample_ids1 = self.categorization_result.index.values
            set_dict = perm_sample_info.to_dict()["Perm_SET"]
            datasets = np.array([set_dict[sample_id] for sample_id in sample_ids1])
            
            sample_ids2 = self.categorization_result[datasets == "test"].index.values
            phenotype_dict = perm_sample_info.to_dict()["Perm_PHENOTYPE"]
            test_response = np.array([phenotype_dict[sample_id] == "case" for sample_id in sample_ids2])
            
            sample_ids3 = self.categorization_result[datasets == "training"].index.values
            response = np.array([phenotype_dict[sample_id] == "case" for sample_id in sample_ids3])
            
            covariates = self.categorization_result[datasets == "training"]
            test_covariates = self.categorization_result[datasets == "test"]
            
            ctrl_var_counts = pd.concat(
                [covariates[filtered_combs][~response],
                 test_covariates[filtered_combs][~test_response]],
            ).sum()
            rare_idx = (ctrl_var_counts < self.ctrl_thres).values
            log.print_progress(f"# of rare categories (Seed: {seed}): {rare_idx.sum()}")
            cov = covariates[filtered_combs].iloc[:, rare_idx]
            test_cov = test_covariates[filtered_combs].iloc[:, rare_idx]
        else:
            response, test_response = self.response, self.test_response
            
            ctrl_var_counts = pd.concat(
                [self.covariates[filtered_combs][~response],
                 self.test_covariates[filtered_combs][~test_response]],
            ).sum()
            rare_idx = (ctrl_var_counts < self.ctrl_thres).values
            log.print_progress(f"# of rare categories (Seed: {seed}): {rare_idx.sum()}")
            cov = self.covariates[filtered_combs].iloc[:, rare_idx]
            test_cov = self.test_covariates[filtered_combs].iloc[:, rare_idx]

        if cov.shape[1] == 0:
            log.print_warn(f"There are no rare categories (Seed: {seed}).")
            return
        
        scaler = StandardScaler().fit(cov)
        cov2 = scaler.transform(cov)
        test_cov2 = scaler.transform(test_cov)

        y = np.where(response, 1.0, 0.0)
        test_y = np.where(test_response, 1.0, 0.0)
        log.print_progress(f"Running LassoCV (Seed: {seed})")
        
        #lasso_model = ElasticNet(alpha=1, n_lambda=100, standardize=True, n_splits=self.fold, n_jobs=self.num_proc,
        #                         scoring='mean_squared_error', random_state=seed)
        
        lasso_model = ElasticNetCV(l1_ratio=1, cv = self.custom_cv_folds(seed=seed), n_jobs = self.num_proc,
                                   random_state = seed, verbose = False, n_alphas=100, selection='random')

        lasso_model.fit(cov2, y)
        #opt_model_idx = np.argmax(getattr(lasso_model, 'cv_mean_score_'))
        #coeffs = getattr(lasso_model, 'coef_path_')
        coeffs = getattr(lasso_model, 'coef_')
        opt_coeff = np.zeros(len(rare_idx))
        opt_coeff[rare_idx] = coeffs
        
        #opt_lambda = getattr(lasso_model, 'lambda_max_')
        opt_lambda = getattr(lasso_model, 'alpha_')
        n_select = np.sum(np.abs(opt_coeff) != 0.0)
        y_pred = lasso_model.predict(test_cov2)
        rsq = r2_score(test_y, y_pred)
        result_dict[domain][seed] = [opt_lambda, rsq, n_select, opt_coeff]
        
        log.print_progress("Done")
            
    def custom_cv_folds(self, seed: int = 42) -> Tuple[np.ndarray, np.ndarray]:
        np.random.seed(seed)
        nobs = np.sum(self.datasets == "training")
        rand_idx = np.random.permutation(nobs)
        foldid = np.repeat(0, nobs)
        i=0
        while i<=self.fold:
            idx_val = rand_idx[np.arange(nobs * (i - 1) / self.fold, nobs * i / self.fold, dtype=int)]
            foldid[idx_val] = i
            i+=1
    
    def permute_pvalues(self):
        """Run LassoCV to get permutated pvalues"""
        log.print_progress(self.permute_pvalues.__doc__)
        
        @contextmanager
        def nullify_output(suppress_stdout: bool=True, suppress_stderr: bool=True):
            stdout = sys.stdout
            stderr = sys.stderr
            devnull = open(os.devnull, "w")
            try:
                if suppress_stdout:
                    sys.stdout = devnull
                if suppress_stderr:
                    sys.stderr = devnull
                yield
            finally:
                if suppress_stdout:
                    sys.stdout = stdout
                if suppress_stderr:
                    sys.stderr = stderr
                    
        seeds = np.arange(self.seed, self.seed+self.n_permute)

        domain_values = [domain.strip() for domain in self.domain_list.split(',')]
        for domain in domain_values:
            if domain == 'all':
                filtered_combs = pd.Series(self.categorization_result.columns)
            else:
                filtered_combs = self.category_set.loc[self.category_set['is_'+domain]==1]['Category']

            for seed in tqdm(seeds):
                with nullify_output():
                    self.risk_score_per_category(domain = domain, result_dict=self._permutation_dict, seed=seed, swap_label=True, filtered_combs=filtered_combs)
                

    def save_results(self):
        """Save the results to a file """
        log.print_progress(self.save_results.__doc__)

        domain_list = list(self._result_dict.keys())
        null_models = []
        fin_res = pd.DataFrame()

        for domain in domain_list:
            if domain == 'all':
                filtered_combs = pd.Series(self.categorization_result.columns)
            else:
                filtered_combs = self.category_set.loc[self.category_set['is_'+domain]==1]['Category']

            result_table = []

            choose_idx = np.all([self._result_dict[domain][seed][3] != 0 
                                for seed in self._result_dict[domain].keys()], axis=0)
            ## Get the categories which are selected by the LassoCV for all seeds
            coef_df = pd.DataFrame.from_dict(
                {seed: self._result_dict[domain][seed][3][choose_idx]
                for seed in self._result_dict[domain].keys()},
                orient="index",
                columns = filtered_combs[choose_idx]
            )

            # Create the initial DataFrame with the 'Domain' column
            #domain_column_data = {'Domain': ['noncoding'] * coef_df.shape[0]}
            #domain_df = pd.DataFrame(domain_column_data)
            #domain_df

            coef_df.to_csv(
                str(self.coef_path).replace('.csv', f'.{domain}.csv'),
                sep="\t"
            )

            for seed in self._result_dict[domain].keys():
                result_table += [[domain] + [str(seed)] + self._result_dict[domain][seed][:-1]]
                #[opt_lambda, rsq, n_select, opt_coeff]
            result_df = pd.DataFrame(result_table, columns=["Domain", "Seed", "Parameter", "R2", "N_select"])
            new_df = pd.DataFrame([domain, 'average', result_df['Parameter'].mean(), result_df['R2'].mean(), sum(choose_idx)]).T
            new_df.columns = result_df.columns
            result_df = pd.concat([new_df, result_df], ignore_index=True)

            if not self.predict_only:
                r2_scores = np.array([self._permutation_dict[domain][seed][1]
                                    for seed in self._permutation_dict[domain].keys()])
                null_models.append([domain, 'average', r2_scores.mean(), r2_scores.std()])
                null_models.extend([[domain] + [i+1] + [str(r2_scores[i])] + [''] for i in range(len(r2_scores))])

                new_values = []
                for row in result_df['R2']:
                    new_value = (np.sum(r2_scores >= row) + 1) / (len(r2_scores) + 1)
                    new_values.append(new_value)

                result_df['Perm_P'] = new_values
                
                self.draw_histogram_plot(domain=domain,
                                         r2 = float(result_df.loc[result_df['Seed'] == 'average']['R2'].values),
                                         perm_r2 = r2_scores)
            
            fin_res = pd.concat([fin_res, result_df], ignore_index=True)

        fin_res.to_csv(self.result_path, sep="\t", index=False)
        if not self.predict_only:
            null_models = pd.DataFrame(
                null_models,
                columns=["Domain", "N_perm", "R2", "std"]
            )
            null_models.to_csv(self.null_model_path, sep="\t", index=False)


    def update_env(self):
        """Update the environment variables """
        log.print_progress(self.update_env.__doc__)
        
        self.set_env("LASSO_RESULTS", self.result_path)
        if not self.predict_only:
            self.set_env("LASSO_NULL_MODELS", self.null_model_path)
        self.save_env()
        
    def draw_histogram_plot(self, domain: str, r2: float, perm_r2: np.ndarray):
        log.print_progress("Save histogram plot")
        
        # Set the font size
        plt.rcParams.update({'font.size': 8})
        
        # Set the figure size
        plt.figure(figsize=(4, 4))

        # Create the histogram plot
        plt.hist(perm_r2, bins=20, color='lightgrey', edgecolor='black')
        
        text_label1 = 'P={:.2f}'.format((sum(perm_r2>=r2)+1)/(len(perm_r2)+1))
        text_label2 = '$R^2$={:.2f}%'.format(r2*100)

        # Add labels and title
        plt.xlabel('$R^2$')
        plt.ylabel('Frequency')
        plt.title('Histogram Plot', fontsize = 8)
        plt.axvline(x=r2, color='red')
        plt.text(0.05, 0.95, text_label1, transform=plt.gca().transAxes, ha='left', va='top', fontsize=8, color='black')
        plt.text(0.05, 0.85, text_label2, transform=plt.gca().transAxes, ha='left', va='top', fontsize=8, color='red')
        plt.locator_params(axis='x', nbins=5)

        plt.savefig(str(self.plot_path).replace('.pdf', f'.{domain}.pdf'))
        






import argparse
import shutil
from pathlib import Path
from typing import Optional

import cwas.utils.log as log
from cwas.runnable import Runnable


class Start(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.config_path = self.workspace / "configuration.txt"

    @staticmethod
    def _create_arg_parser() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            description="Arguments for Initializing a CWAS workspace",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        default_workspace = Path.home() / ".cwas"
        parser.add_argument(
            "-w",
            "--workspace",
            dest="workspace",
            required=False,
            type=Path,
            default=default_workspace,
            help="Path to your CWAS workspace directory",
        )
        return parser

    @staticmethod
    def _print_args(args: argparse.Namespace):
        log.print_arg("CWAS Workspace", args.workspace)

    @property
    def workspace(self):
        return self.args.workspace.resolve()

    def run(self):
        if self.workspace.is_dir():
            log.print_warn(
                f"The workspace '{self.workspace}' already exists. "
                f"This will be not re-created."
            )
        else:
            self._create_workspace()

        self._update_env()

        if self.config_path.is_file():
            log.print_warn(
                f"The configuration file '{self.config_path}' already exists. "
                f"This will be not re-created"
            )
        else:
            self._create_config_file()

    def _create_workspace(self):
        log.print_progress(f"Create CWAS workspace '{self.workspace}'")
        try:
            self.workspace.mkdir(exist_ok=True)
        except NotADirectoryError:
            log.print_err("The path to CWAS workspace is invalid.")
            raise

    def _update_env(self):
        self.env.set_env("CWAS_WORKSPACE", self.workspace)
        self.env.save()

    def _create_config_file(self):
        config = self._init_config()
        with self.config_path.open("w") as config_file:
            for k, v in config.items():
                print(f"{k}={v}", file=config_file)

    def _init_config(self) -> dict:
        config = {
            "ANNOTATION_DATA_DIR": "",
            "GENE_MATRIX": "",
            "ANNOTATION_KEY_CONFIG": "",
            "VEP": "",
            "VEP_CACHE_DIR": "",
            "VEP_CONSERVATION_FILE": "",
            "VEP_LOFTEE": "",
            "VEP_HUMAN_ANCESTOR_FA": "",
            "VEP_GERP_BIGWIG": "",
            "VEP_MIS_DB": "",
            "VEP_MIS_INFO_KEY": "",
            "VEP_MIS_THRES": "",
        }

        installed_vep = self._find_vep_path()
        if installed_vep:
            config["VEP"] = installed_vep

        return config

    def _find_vep_path(self) -> Optional[str]:
        log.print_progress("Find pre-installed VEP")
        vep = shutil.which("vep")
        log.print_progress(f"VEP path: '{vep}'")
        return vep



import pytest
from cwas.core.categorization.category import Category


def test_category_hashable():
    category1 = Category("A1", "B1", "C1", "D1", "E1")
    category2 = Category("A2", "B2", "C2", "D2", "E2")
    category3 = Category("A1", "B1", "C1", "D1", "E1")

    test_dict = {}
    test_dict[category1] = 5
    test_dict[category2] = 2
    test_dict[category3] += 1

    assert test_dict[category1] == 6
    assert test_dict[category2] == 2


def test_category_hashable_keyerror():
    category1 = Category("A1", "B1", "C1", "D1", "E1")
    category2 = Category("A1", "B2", "C1", "D2", "E1")

    test_dict = {}
    test_dict[category1] = 1

    with pytest.raises(KeyError):
        test_dict[category2] += 1


def test_get_category_from_dict():
    category1 = Category("A1", "B1", "C1", "D1", "E1")
    test_dict = {}
    test_dict[category1] = 1

    category_from_dict = list(test_dict.keys())[0]
    assert category1 == category_from_dict




from pathlib import Path
from typing import Optional
import argparse

import numpy as np
import pandas as pd
import re

from collections import defaultdict
from itertools import product

from cwas.runnable import Runnable
from cwas.utils.log import print_progress
from cwas.core.categorization.parser import (
    parse_annotated_vcf)
from cwas.utils.check import check_is_file, check_is_dir

class ExtractVariant(Runnable):
    def __init__(self, args: Optional[argparse.Namespace] = None):
        super().__init__(args)
        self._annotated_vcf = None
        self._gene_matrix = None
        self._tag = None
        self._category_set_path = None
        self._category_set = None

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.input_path)
        check_is_dir(args.output_dir_path)
        if args.category_set_path :
            check_is_file(args.category_set_path)

    @property
    def input_path(self):
        return self.args.input_path.resolve()

    @property
    def output_dir_path(self):
        return self.args.output_dir_path.resolve()

    @property
    def annotation_info(self) -> bool:
        return self.args.annotation_info

    @property
    def tag(self) -> str:
        return self.args.tag

    @property
    def mis_info_key(self) -> str:
        return self.get_env("VEP_MIS_INFO_KEY")

    @property
    def mis_thres(self) -> float:
        return float(self.get_env("VEP_MIS_THRES"))

    @property
    def annotated_vcf(self) -> pd.DataFrame:
        if self._annotated_vcf is None:
            print_progress("Parse the annotated VCF")
            self._annotated_vcf = parse_annotated_vcf(
                Path(self.input_path)
            )
        return self._annotated_vcf

    @property
    def gene_matrix(self) -> pd.DataFrame:
        if self._gene_matrix is None:
            self._gene_matrix = pd.read_csv(self.get_env("GENE_MATRIX"), sep='\t')
            # Keep the last gene in duplicates
            self._gene_matrix = self._gene_matrix[~self._gene_matrix.duplicated(subset=['gene_name'], keep='last')]
        return self._gene_matrix

    @property
    def category_set_path(self) -> Optional[Path]:
        return (
            self.args.category_set_path.resolve()
            if self.args.category_set_path
            else None
        )
    @property
    def category_set(self) -> pd.DataFrame:
        if self._category_set is None and self.category_set_path:
            self._category_set = pd.read_csv(self.category_set_path, sep='\t')
        return self._category_set
    
    @property
    def result_path(self) -> Path:
        if self.tag is None:
            save_name = 'extracted_variants.txt.gz'
        else:
            save_name = '.'.join([self.tag, 'extracted_variants.txt.gz'])
        f_name = re.sub(r'annotated\.vcf\.gz|annotated\.vcf', save_name, self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )
    
    def annotate_variants(self):
        print_progress("Annotate variants with annotation dataset")
        self.annotated_vcf['CLASS'] = np.where(self.annotated_vcf['REF'].str.len() > 1, 'Deletion',
                                               np.where((self.annotated_vcf['REF'].str.len() == 1) & (self.annotated_vcf['ALT'].str.len() == 1), 'SNV', 'Insertion'))
        self.annotated_vcf['DEF.GENE'] = self.annotated_vcf.apply(lambda x: x['NEAREST']
                                                                    if "downstream_gene_variant" in x['Consequence']
                                                                    or "intergenic_variant" in x['Consequence']
                                                                    else x['SYMBOL'], axis=1)
        merged_df = pd.merge(self.annotated_vcf, self.gene_matrix, left_on='DEF.GENE', right_on='gene_name', how='left')
        cols_to_fillna = self.gene_matrix.columns.tolist()
        merged_df[cols_to_fillna] = merged_df[cols_to_fillna].fillna(0)
        merged_df = merged_df.drop(columns=['gene_id', 'gene_name'])
        ## Coding
        # Define the list of string patterns to search for
        patterns = ['stop_gained', 'splice_donor', 'splice_acceptor', 'frameshift_variant', 'missense_variant', 'protein_altering_variant', 'start_lost', 'stop_lost', 'inframe_deletion', 'inframe_insertion', 'synonymous_variant', 'stop_retained_variant', 'incomplete_terminal_codon_variant', 'protein_altering_variant', 'coding_sequence_variant']
        # Use str.contains() to search for each pattern in the Consequence column
        merged_df['is_CodingRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_CodingRegion'] = np.where(merged_df['ProteinCoding'] == 1, merged_df['is_CodingRegion'], 0)
        ## Noncoding
        merged_df['is_NoncodingRegion'] = np.where(merged_df['is_CodingRegion'] == 1, 0, 1)
        ## PTV
        # Define the list of string patterns to search for
        patterns = ['stop_gained', 'splice_donor', 'splice_acceptor', 'frameshift_variant']
        # Use str.contains() to search for each pattern in the Consequence column
        merged_df['is_PTVRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_PTVRegion'] = np.where((merged_df['is_CodingRegion'] == 1) & (merged_df['LoF'] == 'HC') & ((merged_df['LoF_flags'] == 'SINGLE_EXON') | (merged_df['LoF_flags'] == '')), merged_df['is_PTVRegion'], 0)
        ## Frameshift
        # Define the list of string patterns to search for
        patterns = ['frameshift_variant']
        # Define the list of string patterns to exclude
        exclude_patterns = ['stop_gained', 'splice_donor', 'splice_acceptor']
        merged_df['is_FrameshiftRegion'] = ((merged_df['Consequence'].str.contains('|'.join(patterns))) 
                                            & (~merged_df['Consequence'].str.contains('|'.join(exclude_patterns)))).astype(int)
        merged_df['is_FrameshiftRegion'] = np.where(merged_df['is_PTVRegion'] == 1, merged_df['is_FrameshiftRegion'], 0)
        ## Missense
        patterns = ['missense_variant', 'protein_altering_variant', 'start_lost', 'stop_lost']
        merged_df['is_MissenseRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_MissenseRegion'] = np.where((merged_df['is_CodingRegion'] == 1) & (merged_df['is_PTVRegion'] == 0), merged_df['is_MissenseRegion'], 0)
        ## Damaging missense
        merged_df['is_DamagingMissenseRegion'] = np.where((merged_df['is_MissenseRegion'] == 1) & (pd.to_numeric(merged_df["MisDb_" + self.mis_info_key], errors='coerce').fillna(0) >= self.mis_thres), 1, 0)
        # Define the list of string patterns to search for
        patterns = ['inframe_deletion', 'inframe_insertion']
        merged_df['is_InFrameRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_InFrameRegion'] = np.where((merged_df['is_CodingRegion'] == 1) & (merged_df['is_PTVRegion'] == 0) & (merged_df['is_MissenseRegion'] == 0), merged_df['is_InFrameRegion'], 0)
        ## Silent
        patterns = ['synonymous_variant']
        merged_df['is_SilentRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_SilentRegion'] = np.where((merged_df['is_CodingRegion'] == 1) &
                                                (merged_df['is_PTVRegion'] == 0) &
                                                (merged_df['is_MissenseRegion'] == 0) &
                                                (merged_df['is_InFrameRegion'] == 0),
                                                merged_df['is_SilentRegion'], 0)
        ## UTRs
        # Define the list of string patterns to search for
        patterns = ['_UTR_']
        merged_df['is_UTRsRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_UTRsRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1),
                                              merged_df['is_UTRsRegion'], 0)
        ## Promoter
        patterns = ['upstream_gene_variant']
        merged_df['is_PromoterRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_PromoterRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0),
                                                  merged_df['is_PromoterRegion'], 0)
        ## Intron
        patterns = ['intron_variant']
        merged_df['is_IntronRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_IntronRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0) &
                                                  (merged_df['is_PromoterRegion'] == 0),
                                                  merged_df['is_IntronRegion'], 0)
        ## SpliceSite non-canonical
        patterns = ['splice_region_variant']
        merged_df['is_SpliceSiteNoncanonRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_SpliceSiteNoncanonRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0) &
                                                  (merged_df['is_PromoterRegion'] == 0) &
                                                  (merged_df['is_IntronRegion'] == 0),
                                                  merged_df['is_SpliceSiteNoncanonRegion'], 0)
        ## Intergenic
        patterns = ['downstream_gene_variant', 'intergenic_variant']
        merged_df['is_IntergenicRegion'] = merged_df['Consequence'].str.contains('|'.join(patterns)).astype(int)
        merged_df['is_IntergenicRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0) &
                                                  (merged_df['is_PromoterRegion'] == 0) &
                                                  (merged_df['is_IntronRegion'] == 0) &
                                                  (merged_df['is_SpliceSiteNoncanonRegion'] == 0),
                                                  merged_df['is_IntergenicRegion'], 0)
        ## lincRNA
        merged_df['is_lincRnaRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0) &
                                                  (merged_df['is_PromoterRegion'] == 0) &
                                                  (merged_df['is_IntronRegion'] == 0) &
                                                  (merged_df['is_SpliceSiteNoncanonRegion'] == 0) &
                                                  (merged_df['is_IntergenicRegion'] == 0) &
                                                  (merged_df['lincRNA'] == 1) &
                                                  (merged_df['ProteinCoding'] == 0),
                                                  1, 0)
        ## Others
        merged_df['is_OtherTranscriptRegion'] = np.where((merged_df['is_NoncodingRegion'] == 1) &
                                                  (merged_df['is_UTRsRegion'] == 0) &
                                                  (merged_df['is_PromoterRegion'] == 0) &
                                                  (merged_df['is_IntronRegion'] == 0) &
                                                  (merged_df['is_SpliceSiteNoncanonRegion'] == 0) &
                                                  (merged_df['is_IntergenicRegion'] == 0) &
                                                  (merged_df['is_lincRnaRegion'] == 0) &
                                                  (merged_df['ProteinCoding'] == 0),
                                                  1, 0)
        self._result = merged_df

    def allocate_variants(self, category: pd.DataFrame):
        if category['variant_type'] == 'All':
            filtered_result = self._result[self._result['CLASS'].isin(['SNV', 'Deletion', 'Insertion'])]
        elif category['variant_type'] == 'SNV':
            filtered_result = self._result.query('CLASS == "SNV"')
        else:
            filtered_result = self._result[self._result['CLASS'].isin(['Deletion', 'Insertion'])]

        if category['gene_set'] != 'Any':
            filtered_result = filtered_result[filtered_result[category['gene_set']] == 1]
        if category['functional_score'] != 'All':
            filtered_result = filtered_result[filtered_result[category['functional_score']] == 1]
        if category['gencode'] != 'Any':
            filtered_result = filtered_result[filtered_result['_'.join(['is', category['gencode']])] == 1]
        if category['functional_annotation'] != 'Any':
            filtered_result = filtered_result[filtered_result[category['functional_annotation']] == 1]
        filtered_result['CATEGORY'] = category['Category']
    
        return(filtered_result)

    def filter_variants(self):
        print_progress(f"Filter variants in {self.category_set.shape[0]} categories")
        self.category_set[['variant_type', 'gene_set', 'functional_score', 'gencode', 'functional_annotation']] = self.category_set['Category'].str.split('_', expand=True)
        # Filter variants by categories and concatenate them vertically
        self._result = pd.concat(self.category_set.apply(lambda x: self.allocate_variants(category = x), axis=1).tolist(), axis=0)
    
    def remove_annotation_info(self):
        print_progress("No annotation information attached")
        if self.category_set_path :
            self._result = self._result.loc[:, ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'SAMPLE', 'CATEGORY']]
        else :
            self._result = self._result.loc[:, ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'SAMPLE']]
    
    def save_result(self):
        print_progress(f"Save the result to the file {self.result_path}")
        self._result.to_csv(self.result_path, sep='\t', compression='gzip', index=False)
    
    def run(self):
        self.annotate_variants()
        if self.category_set_path :
            self.filter_variants()
        if self.annotation_info is None :
            self.remove_annotation_info()
        else:
            print_progress("Annotation information attached")
        self.save_result()
        print_progress("Done")
        
        


"""
Module for the categorization of the variants into CWAS categories and
counting the number of variants in each category.
CWAS category is a combination of annotation terms
from each of annotation groups.

There are currently 5 groups of annotation terms.

--- The groups of the annotation terms for CWAS ---
    1. Variant types (variant_type)
    2. Functional score (functional_score)
    3. Gene lists (gene_set)
    4. GENCODE annotation categories (gencode)
    5. Functional annotation categories (functional_annotation)

"""
from collections import defaultdict
from itertools import product

import numpy as np
import pandas as pd
from tqdm import tqdm
from cwas.core.categorization.utils import extract_sublist_by_int, get_idx_dict


class Categorizer:
    def __init__(self, category_domain: dict, gene_matrix: dict, mis_info_key: str, mis_thres: float) -> None:
        self._category_domain = category_domain
        self._gene_matrix = gene_matrix
        self._mis_info_key = mis_info_key
        self._mis_thres = mis_thres

    def categorize_variant(self, annotated_vcf: pd.DataFrame):
        result = defaultdict(int)

        for annotation_term_lists in self.annotate_each_variant(annotated_vcf):
            for combination in product(*annotation_term_lists):
                result["_".join(combination)] += 1

        return result

    def get_intersection(self, annotated_vcf: pd.DataFrame):
        result = defaultdict(lambda: defaultdict(int))

        for annotation_term_lists in tqdm(self.annotate_each_variant(annotated_vcf), total=len(annotated_vcf)):
            for combination in product(*annotation_term_lists):
                for combination2 in product(*annotation_term_lists):
                    result["_".join(combination)]["_".join(combination2)] += 1

        return result

    def annotate_each_variant(self, annotated_vcf):
        """Newly annotated each variant using CWAS annotation terms.
        In order to annotate each variant with multiple annotation terms
        from each group efficiently, "Annotation integer" has defined.

        Annotation integer: A bitwise representation of the annotation
        of each variant where each bit means each annotation term

        e.g. If the annotation terms is ['A', 'B', 'C', 'D'] and
        the annotation integer is 0b1011, it means that
        the variant is annotated as 'A', 'B' and 'D'.
        """
        variant_type_annotation_ints = self.annotate_variant_type(annotated_vcf)
        functional_score_annotation_ints = self.annotate_functional_score(annotated_vcf)
        gene_set_annotation_ints = self.annotate_gene_set(annotated_vcf)
        gencode_annotation_ints = self.annotate_gencode(annotated_vcf)
        region_annotation_ints = self.annotate_region(annotated_vcf)

        for (
            variant_type_annotation_int,
            functional_score_annotation_int,
            gene_set_annotation_int,
            gencode_annotation_int,
            region_annotation_int,
        ) in zip(
            variant_type_annotation_ints,
            functional_score_annotation_ints,
            gene_set_annotation_ints,
            gencode_annotation_ints,
            region_annotation_ints,
        ):
            yield (
                self.parse_annotation_int(
                    variant_type_annotation_int, "variant_type",
                ),
                self.parse_annotation_int(
                    gene_set_annotation_int, "gene_set"
                ),
                self.parse_annotation_int(
                    functional_score_annotation_int, "functional_score",
                ),
                self.parse_annotation_int(gencode_annotation_int, "gencode"),
                self.parse_annotation_int(region_annotation_int, "functional_annotation"),
            )

    def annotate_variant_type(self, annotated_vcf: pd.DataFrame) -> list:
        variant_type_annotation_idx = get_idx_dict(
            self._category_domain["variant_type"]
        )
        refs = annotated_vcf["REF"].values
        alts = annotated_vcf["ALT"].values

        is_snv_arr = (
            (np.vectorize(len)(refs) == 1) & (np.vectorize(len)(alts) == 1)
        ).astype(np.int32)
        annotation_int_conv = (
            lambda is_snv: 2 ** variant_type_annotation_idx["SNV"]
            if is_snv
            else 2 ** variant_type_annotation_idx["Indel"]
        )
        annotation_ints = np.vectorize(annotation_int_conv)(is_snv_arr)
        annotation_ints += 2 ** variant_type_annotation_idx["All"]

        return annotation_ints

    def annotate_functional_score(self, annotated_vcf: pd.DataFrame) -> list:
        functional_score_annotation_idx = get_idx_dict(
            self._category_domain["functional_score"]
        )
        annotation_ints = np.zeros(len(annotated_vcf.index), dtype=int)

        for score in functional_score_annotation_idx:
            if score == "All":
                continue

            score_vals = annotated_vcf[score].values.astype(np.int32)
            annotation_int_conv_func = (
                lambda x: 2 ** functional_score_annotation_idx[score] * x
            )
            annotation_ints += np.vectorize(annotation_int_conv_func)(
                score_vals
            )

        annotation_ints += 2 ** functional_score_annotation_idx["All"]

        return annotation_ints

    def annotate_gene_set(self, annotated_vcf: pd.DataFrame) -> list:
        gene_set_annotation_idx = get_idx_dict(
            self._category_domain["gene_set"]
        )
        gene_symbols = annotated_vcf["SYMBOL"].values
        gene_nearests = annotated_vcf["NEAREST"].values
        gencodes = annotated_vcf["Consequence"].values  # GENCODE annotations

        annotation_int_list = []
        annotation_int_dict = {}

        for symbol, nearest, gencode in zip(
            gene_symbols, gene_nearests, gencodes
        ):
            gene = (
                nearest
                if "downstream_gene_variant" in gencode
                or "intergenic_variant" in gencode
                else symbol
            )
            annotation_int = annotation_int_dict.get(gene, 0)

            if annotation_int == 0:
                gene_set_ = self._gene_matrix.get(gene, set())

                if gene_set_:
                    for gene_cat in gene_set_annotation_idx:
                        if gene_cat in gene_set_:
                            annotation_int += (
                                2 ** gene_set_annotation_idx[gene_cat]
                            )

            annotation_int_list.append(annotation_int)

        annotation_ints = np.asarray(annotation_int_list)
        annotation_ints += 2 ** gene_set_annotation_idx["Any"]

        return annotation_ints

    def annotate_gencode(self, annotated_vcf: pd.DataFrame) -> list:
        gencode_annotation_idx = get_idx_dict(self._category_domain["gencode"])
        gene_symbols = annotated_vcf["SYMBOL"].values
        gene_nearests = annotated_vcf["NEAREST"].values
        gencodes = annotated_vcf["Consequence"].values
        lofs = annotated_vcf["LoF"].values
        lof_flags = annotated_vcf["LoF_flags"].values
        mis_scores = annotated_vcf["MisDb_" + self._mis_info_key].values

        annotation_int_list = []

        for symbol, nearest, gencode, lof, lof_flag, mis_score in zip(
            gene_symbols, gene_nearests, gencodes, lofs, lof_flags, mis_scores
        ):
            gene = (
                nearest
                if "downstream_gene_variant" in gencode
                or "intergenic_variant" in gencode
                else symbol
            )
            gene_set_ = self._gene_matrix.get(gene, set())
            annotation_int = 0
            is_in_coding = False

            if "ProteinCoding" in gene_set_:
                is_in_coding = True
                annotation_int += 2 ** gencode_annotation_idx["CodingRegion"]

                # Coding region
                if ((
                    "stop_gained" in gencode
                    or "splice_donor" in gencode
                    or "splice_acceptor" in gencode
                )
                and (lof == "HC")
                and ((lof_flag=='SINGLE_EXON') or (lof_flag==""))
                ):
                    annotation_int += 2 ** gencode_annotation_idx["PTVRegion"]
                elif ((
                    "frameshift_variant" in gencode
                )
                and (lof == "HC")
                and ((lof_flag=='SINGLE_EXON') or (lof_flag==""))
                ):
                    annotation_int += 2 ** gencode_annotation_idx["PTVRegion"]
                    annotation_int += (
                        2 ** gencode_annotation_idx["FrameshiftRegion"]
                    )
                elif (
                    "missense_variant" in gencode
                    or "protein_altering_variant" in gencode
                    or "start_lost" in gencode
                    or "stop_lost" in gencode
                ):
                    annotation_int += (
                        2 ** gencode_annotation_idx["MissenseRegion"]
                    )

                    if ((len(mis_score)!=0)
                        and (mis_score!='NA')):
                        if '&' in mis_score:
                            mis_score = max([float(i) for i in mis_score.split('&')])
                        else:
                            mis_score = float(mis_score)
                        if mis_score >= self._mis_thres:
                            annotation_int += (
                                2 ** gencode_annotation_idx["DamagingMissenseRegion"]
                            )

                elif (
                    "inframe_deletion" in gencode
                    or "inframe_insertion" in gencode
                ):
                    annotation_int += (
                        2 ** gencode_annotation_idx["InFrameRegion"]
                    )
                elif "synonymous_variant" in gencode:
                    annotation_int += (
                        2 ** gencode_annotation_idx["SilentRegion"]
                    )
                elif (
                    "stop_retained_variant" not in gencode
                    and "incomplete_terminal_codon_variant" not in gencode
                    and "protein_altering_variant" not in gencode
                    and "coding_sequence_variant" not in gencode
                    and "stop_gained" not in gencode
                    and "splice_donor" not in gencode
                    and "splice_acceptor" not in gencode
                    and "frameshift_variant" not in gencode
                ):
                    # Noncoding
                    annotation_int = 0
                    is_in_coding = False

            if not is_in_coding:
                annotation_int += 2 ** gencode_annotation_idx["NoncodingRegion"]

                if "_UTR_" in gencode:
                    annotation_int += 2 ** gencode_annotation_idx["UTRsRegion"]
                elif "upstream_gene_variant" in gencode:
                    annotation_int += (
                        2 ** gencode_annotation_idx["PromoterRegion"]
                    )
                elif "intron_variant" in gencode:
                    annotation_int += (
                        2 ** gencode_annotation_idx["IntronRegion"]
                    )
                elif "splice_region_variant" in gencode:
                    annotation_int += (
                        2 ** gencode_annotation_idx["SpliceSiteNoncanonRegion"]
                    )
                elif (
                    "downstream_gene_variant" in gencode
                    or "intergenic_variant" in gencode
                ):
                    annotation_int += (
                        2 ** gencode_annotation_idx["IntergenicRegion"]
                    )
                elif "ProteinCoding" not in gene_set_:
                    if "lincRNA" in gene_set_:
                        annotation_int += (
                            2 ** gencode_annotation_idx["lincRnaRegion"]
                        )
                    else:
                        annotation_int += (
                            2 ** gencode_annotation_idx["OtherTranscriptRegion"]
                        )

            annotation_int_list.append(annotation_int)

        annotation_ints = np.asarray(annotation_int_list)
        annotation_ints += 2 ** gencode_annotation_idx["Any"]

        return annotation_ints

    def annotate_region(self, annotated_vcf: pd.DataFrame) -> list:
        region_annotation_idx = get_idx_dict(self._category_domain["functional_annotation"])
        annotation_floats = np.zeros(len(annotated_vcf.index), dtype=float)

        for region in region_annotation_idx:
            if region == "Any":
                continue

            region_vals = annotated_vcf[region].values.astype(np.int32)
            annotation_int_conv_func = (
                lambda x: 2 ** region_annotation_idx[region] * x
            )
            annotation_floats += np.vectorize(annotation_int_conv_func, otypes=[float])(
                region_vals
            )

        annotation_floats += 2 ** region_annotation_idx["Any"]
        annotation_ints = np.array([int(x) for x in annotation_floats])

        return annotation_ints

    def parse_annotation_int(
        self, annotation_int: int, annotation_term_type: str
    ) -> list:
        """ Parse the annotation integer and
        choose the appropriate subset from the specific annotation terms.
        """
        return extract_sublist_by_int(
            self._category_domain[annotation_term_type], annotation_int
        )



#!/usr/bin/env python3
"""Category-wide association study (CWAS)

This is a data analytic tool to perform stringent association tests
to find non-coding loci associated with autism spectrum disorder (ASD).
"""
import setuptools

setuptools.setup(
    name="cwas",
    version="0.1.0",
    license="MIT",
    author="Minwoo Jeong",
    author_email="jeongmwj@gmail.com",
    description=__doc__,
    url="https://github.com/joonan-lab/cwas",
    project_urls={
        "Bug Tracker": "https://github.com/joonan-lab/cwas/issues",
    },
    packages=setuptools.find_packages(),
    package_data={'': ['*.r', '*.R']},
    include_package_data=True,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.7',
    entry_points={
        'console_scripts': [
            'cwas = cwas.__main__:main'
        ]
    },
    install_requires=["pandas", "scipy", "tqdm", "igraph", "matplotlib", "rpy2", "seaborn", "parmap",
                      "polars", "scikit-learn", "pysam", "pytabix", "pyyaml", "python-dotenv", "pytest", "pyarrow",
                      "adjustText", "numpy"]
)



"""
Test the 'BedReader' class in the cwas.core.preparation.bedreader module
"""
from pathlib import Path

import pytest
from cwas.core.preparation.bedreader import BedReader
from cwas.utils.cmd import compress_using_bgzip, index_using_tabix


@pytest.fixture(scope="module")
def bed_coordinates():
    coordinates = [
        ("chr1", 1, 2, "A"),
        ("chr2", 20, 30, "B"),
        ("chr3", 100, 200, "C"),
    ]
    return coordinates


@pytest.fixture(scope="module")
def bed_txt_path(cwas_workspace):
    return cwas_workspace / ".test.bed"


@pytest.fixture(scope="module")
def bed_gz_path(bed_txt_path):
    return Path(str(bed_txt_path) + ".gz")


def create_bed_file(bed_txt_path, bed_coordinates):
    with bed_txt_path.open("w") as bed_file:
        for bed_coordinate in bed_coordinates:
            print(*bed_coordinate, sep="\t", file=bed_file)

    bed_gz_path = compress_using_bgzip(bed_txt_path)
    _ = index_using_tabix(bed_gz_path)


@pytest.fixture(scope="module", autouse=True)
def setup(cwas_workspace, bed_txt_path, bed_coordinates):
    cwas_workspace.mkdir()
    create_bed_file(bed_txt_path, bed_coordinates)


def create_bed_file(bed_txt_path, bed_coordinates):
    with bed_txt_path.open("w") as bed_file:
        for bed_coordinate in bed_coordinates:
            print(*bed_coordinate, sep="\t", file=bed_file)

    bed_gz_path = compress_using_bgzip(bed_txt_path)
    _ = index_using_tabix(bed_gz_path)


@pytest.fixture(scope="module", autouse=True)
def teardown(cwas_workspace):
    yield
    remove_workspace(cwas_workspace)


def remove_workspace(cwas_workspace):
    for f in cwas_workspace.glob("*"):
        f.unlink()
    cwas_workspace.rmdir()


def test_init_bedreader(bed_gz_path):
    bedreader = BedReader(bed_gz_path)
    assert isinstance(bedreader, BedReader)


def test_fail_init_bedreader(cwas_workspace):
    invalid_bed_path = cwas_workspace / ".not_exist.bed"
    with pytest.raises(OSError):
        BedReader(invalid_bed_path)


def test_read_coordinates(bed_gz_path, bed_coordinates):
    bedreader = BedReader(bed_gz_path)
    expected = bed_coordinates
    result = list(bedreader)
    assert expected == result


def test_read_coordinates_with_set_contig(bed_gz_path, bed_coordinates):
    contig = "chr1"
    bedreader = BedReader(bed_gz_path)
    bedreader.set_contig(contig)
    expected = [
        bed_coordinate
        for bed_coordinate in bed_coordinates
        if bed_coordinate[0] == contig
    ]
    result = list(bedreader)
    assert expected == result


def test_read_coordinates_with_invalid_contig(bed_gz_path):
    invalid_contig = "test"
    bedreader = BedReader(bed_gz_path)
    bedreader.set_contig(invalid_contig)
    assert list(bedreader) == []



import argparse
import pathlib
import random
import unittest
from unittest.mock import patch

from cwas.runnable import Runnable


class RunnableMock(Runnable):
    @staticmethod
    def _create_arg_parser() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        parser.add_argument("-d", dest="test_dir", type=pathlib.Path)
        parser.add_argument("-f", dest="test_file", type=pathlib.Path)
        parser.add_argument("-n", dest="test_int", type=int)
        return parser

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        pass

    @staticmethod
    def _print_args(args: argparse.Namespace):
        pass

    def run(self):
        pass


class TestRunnable(unittest.TestCase):
    # Enable the Runnable class to be instantiated
    # by assign an empty set to Runnable.__abstractmethods__
    @patch.multiple(Runnable, __abstractmethods__=set())
    def test_get_instance(self):
        inst = Runnable.get_instance(argv=list())
        assert isinstance(inst, Runnable)

    @patch.multiple(Runnable, __abstractmethods__=set())
    def test_run(self):
        inst = Runnable.get_instance(argv=list())
        inst.run()
        assert 1

    def test_argument_assignment(self):
        randint = random.randint(1, 1000000)
        test_dir = pathlib.Path.home() / f".cwas-runnable-test-{randint}"
        test_dir.mkdir()
        test_file = test_dir / "test.txt"
        test_file.touch()

        argv = ["-d", str(test_dir), "-f", str(test_file), "-n", str(randint)]
        inst = RunnableMock.get_instance(argv)
        assert hasattr(inst, "args")

        # Teardown
        test_file.unlink()
        test_dir.rmdir()



from sklearn.metrics import silhouette_score
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import importr
from scipy.spatial.distance import pdist, squareform
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

pandas2ri.activate()

class kmeans_cluster:
    def __init__(self, tsne_out: pd.DataFrame, seed: int) -> None:
                 #output_name: str, k_range: str, k=None) -> None:
        
        self._tsne_out = tsne_out
        self._seed = seed
        self.kmeans_r = importr('stats').kmeans
        self.set_seed = robjects.r('set.seed')

    @property
    def tsne_out(self) -> pd.DataFrame:
        return self._tsne_out
    
    @property
    def seed(self) -> str:
        return self._seed
    
    def optimal_k(self, k_range, output_name):
        start, end = list(map(int, k_range.replace(" ", "").split(",")))
        k_values = list(range(start, end+1))

        self.set_seed(self.seed)

        avg_sil_values = list(map(self._avg_sil, k_values))
        df = pd.DataFrame({"k_values": k_values, "avg_sil_values": avg_sil_values})
        self.avg_sil_df = df

        self._print_silhouett_plot(output_name)
        opt_k = df.loc[df.avg_sil_values==max(df.avg_sil_values), "k_values"].values[0]

        return opt_k
    
    def _avg_sil(self, k):
        pandas2ri.activate()
        km_res = self.kmeans_r(np.array(self.tsne_out), centers=k, nstart=300, iter_max=100)
        km_res_dict = dict(zip(km_res.names, list(km_res)))
        distance_matrix = squareform(pdist(self.tsne_out))
        ss = silhouette_score(distance_matrix, km_res_dict['cluster'], metric='precomputed')

        return ss
    
    def _print_silhouett_plot(self, output_name):
        k_start = min(self.avg_sil_df['k_values'])
        k_end = max(self.avg_sil_df['k_values'])
        max_ss = max(self.avg_sil_df.avg_sil_values)
        k = self.avg_sil_df.loc[self.avg_sil_df.avg_sil_values==max_ss, "k_values"].values[0]
        
        plt.figure(figsize=(20, 10))
        plt.title('Range of K: {} to {}'.format(k_start, k_end), fontsize=15, pad=10)
        plt.plot(self.avg_sil_df['k_values'], self.avg_sil_df['avg_sil_values'], '-o', c="black", linewidth=1, markersize=4)
        plt.vlines(k, 0, max_ss, color='red', linestyles='--')
        plt.text(k, max_ss+0.01, "Optimal K = {}".format(str(k)), fontdict={'color': 'red', 'weight': 'bold', 'fontsize': 15})
        plt.xticks(np.arange(0, int(np.ceil(k_end/10)*10)+1, 10), fontsize=10)
        plt.yticks(np.arange(0, max_ss+0.1, 0.1), fontsize=10)
        plt.xlim(k_start-5, k_end+5)
        plt.ylim(0, max_ss+0.1)
        plt.xlabel("Number of clusters K", fontsize=12, labelpad=10)
        plt.ylabel("Average Silhouettes", fontsize=12, labelpad=10)
        plt.tight_layout()
        plt.savefig(output_name) # final output
        #plt.savefig(self._output_name)


    def center_init(self, k):
        self.set_seed(self.seed)

        km_tsne = self.kmeans_r(self.tsne_out, centers=k, nstart=300, iter_max=100)
        km_tsne_dict = dict(zip(km_tsne.names, list(km_tsne)))

        km_tsne_cluster = [x - 1 for x in list(km_tsne_dict['cluster'])]
        km_tsne_centers = pd.DataFrame(km_tsne_dict['centers'][km_tsne_cluster], columns=["t-SNE1", "t-SNE2"])
        dist_to_center_tsne = np.sqrt(np.sum((self.tsne_out - km_tsne_centers)**2, axis=1))

        self.km_tsne_cluster_ = km_tsne_cluster
        self.dist_to_center_tsne_ = dist_to_center_tsne
        
        i_init_pt = list(map(self._init_pt_kmeans, list(range(k))))

        return i_init_pt

    def _init_pt_kmeans(self, k):
        i_check = list(np.where([x == k for x in self.km_tsne_cluster_])[0])
        i_check_min = self.dist_to_center_tsne_[i_check].idxmin()

        return i_check_min







import argparse, os
from pathlib import Path
import numpy as np
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import seaborn as sns
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.ticker as mticker
import polars as pl
import re

from cwas.utils.log import print_progress, print_arg
from cwas.runnable import Runnable
from cwas.utils.check import check_is_file, check_is_dir

pd.set_option('mode.chained_assignment',  None)

class BurdenShift(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._burden_res = None
        self._burden_shift_res = None
        self._cat_sets = None
        self._cat_counts = None
        self._cat_set_list = None
        
    @staticmethod
    def _print_args(args: argparse.Namespace):
        print_arg("Input file (burden test)", os.path.basename(args.input_path))
        print_arg("Burden shift result file", os.path.basename(args.burden_res))
        print_arg("Output directory", args.output_dir_path)
        print_arg("Category set file", os.path.basename(args.cat_set_file))
        print_arg("Category counts file", os.path.basename(args.cat_count_file))
        print_arg("Cutoff of category sets and p-value", [args.count_cutoff, args.pval])
        print_arg("Output tag (prefix of output files)", args.tag)

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.input_path)
        check_is_file(args.burden_res)
        check_is_file(args.cat_set_file)
        check_is_file(args.cat_count_file)
        check_is_dir(args.output_dir_path)
        
    @property
    def input_file(self) -> Path:
        return self.args.input_path.resolve()

    @property
    def plot_title(self) -> float:
        return self.args.plot_title
    
    @property
    def burden_res(self):
        if self._burden_res is None:
            self._burden_res = pl.read_csv(self.input_file, separator="\t")
        return self._burden_res
    
    @property
    def burden_shift_res(self) -> Path:
        if self._burden_shift_res is None:
            self._burden_shift_res = pl.read_csv(self.args.burden_res.resolve(), separator="\t")
        return self._burden_shift_res       
    
    @property
    def output_dir_path(self) -> Path:
        return self.args.output_dir_path.resolve()
    
    @property
    def cat_set_file(self) -> Path:
        return self.args.cat_set_file.resolve() 

    @property
    def cat_sets(self):
        if self._cat_sets is None:
            self._cat_sets = self._create_category_sets()
        return self._cat_sets
    
    @property
    def cat_counts(self) -> Path:
        if self._cat_counts is None:
            self._cat_counts = pd.read_csv(self.args.cat_count_file.resolve(), sep='\t')
        return self._cat_counts
    
    @property
    def tag(self) -> str:
        return self.args.tag
    
    @property
    def c_cutoff(self) -> int:
        if self.args.count_cutoff < 0:
            raise ValueError(
                print("Count cutoff is must be positive value.")
            )        
        return self.args.count_cutoff
    
    @property
    def pval(self) -> float:
        return self.args.pval
    
    @property
    def cat_set_list(self):
        if not self.args.cat_set_list is None:
            if self._cat_set_list is None:
                with open(self.args.cat_set_list, 'r') as f:
                    self._cat_set_list = f.read().splitlines()
        return self._cat_set_list

    @property
    def n_cat_sets(self) -> int:
        return self.args.n_cat_sets
    
    @property
    def fontsize(self) -> int:
        return self.args.fontsize
    

    def _create_category_sets(self):
        print_progress("Create category sets combined all of GENCODE, regions, and gene list")
        
        catsets = pd.read_csv(self.cat_set_file, sep="\t")
        catsets_dict = catsets.to_dict('list')
        
        genesets = sorted(list(set(catsets['gene_set'].unique()) - set(["Any"])))
        gencodes = ['coding','noncoding','promoter','UTR','intergenic','intron','lincRNA']

        ## all gencodes
        for b in gencodes:
            col = f'is_{b}'
            catsets_dict[col] = list((catsets[f'is_{b}'] == 1).astype(int))

        ## for all genesets
        for g in genesets:
            col = f'is_{g}'
            catsets_dict[col] = list((catsets['gene_set'] == g).astype(int))
            
            # all genesets & all gencodes
            for b in gencodes:
                #if b != 'coding':
                col = f'is_{b}_{g}'
                catsets_dict[col] = list(((catsets['gene_set']==g)&(catsets[f'is_{b}']==1)).astype(int))
            
        ## for all functional_annotations
        regions = sorted(list(set(catsets['functional_annotation'].unique()) - set(["Any"])))
        for r in regions:
            col = f'is_{r}'
            catsets_dict[col] = list((catsets['functional_annotation'] == r).astype(int))
            
            ## all regions
            for b in gencodes:
                #if 'lincRNA' != b:
                col = f'is_{b}_{r}'
                catsets_dict[col] = list(((catsets['functional_annotation'] == r)&(catsets[f'is_{b}'] == 1)).astype(int))

            ## all regions & all genesets
            for g in genesets:
                col = f'is_{r}_{g}'
                catsets_dict[col] = list(((catsets['functional_annotation'] == r)&(catsets['gene_set'] == g)).astype(int))

        ## for all conserved
        functional_score = sorted(list(set(catsets['functional_score'].unique()) - set(["All"])))
        for c in functional_score:
            col = f'is_{c}'
            catsets_dict[col] = list((catsets['functional_score'] == c).astype(int))

            ## all conserved & all gencodes
            for b in gencodes:
                #if 'lincRNA' != b:
                col = f'is_{b}_{c}'
                catsets_dict[col] = list(((catsets['functional_score'] == c)&(catsets[f'is_{b}'] == 1)).astype(int))

            ## all conserved & all genesets
            for g in genesets:
                col = f'is_{r}_{g}'
                catsets_dict[col] = list(((catsets['functional_score'] == c)&(catsets['gene_set'] == g)).astype(int))

        for key in ['variant_type', 'gene_set', 'functional_score', 'gencode', 'functional_annotation']:
            del catsets_dict[key]
            
        _cat_sets = pd.DataFrame(catsets_dict)
        return _cat_sets
    
    def _count_cats(self, pvals, pvalTrash):
        pvals = np.array(pvals)
        nCase = sum((pvals>0) & (pvals<=pvalTrash))
        nCtrl = sum((pvals<0) & (abs(pvals)<=pvalTrash))
        #nCase = len(pvals[(pvals>0) & (pvals<=pvalTrash)])
        #nCtrl = len(pvals[(pvals>0) & (abs(pvals)<=pvalTrash)])
        return (nCase, nCtrl)
    
    def _draw_shiftDistPlot(self, df, setName, nObsCase, nObsCtrl, pCase, pCtrl):
        ggpermCounts = pd.DataFrame(df['case'].tolist()+df['control'].tolist(), columns=['N_signif_tests'])

        ft_size = self.fontsize
        larger = max(nObsCase, nObsCtrl)
        max_cnts = round(ggpermCounts['N_signif_tests'].max())
        xticks = np.arange(max_cnts/4, (max_cnts/4)*4+1, max_cnts/4) if max_cnts > 0 else [max_cnts]
        
        plot_title = setName.replace('is_', '').capitalize()
        subtitle = f'No. cats in Cases: {nObsCase}, No. cats in Controls: {nObsCtrl}'
        
        fig1 = plt.figure(figsize=(5,4))
        plt.title(plot_title, fontsize=ft_size, loc='left', weight='bold')
        plt.suptitle(subtitle, fontsize=ft_size, y=0.92)
        ax = sns.kdeplot(ggpermCounts['N_signif_tests'], edgecolor='black', color='#EBEBEB',
                         alpha=1, fill=True, linewidth=1, warn_singular=False)
        ymax = ax.get_ylim()[-1]
        plt.vlines(ymin=0, ymax=ymax, x=nObsCase, color='red', linestyles='--', linewidth=1)
        plt.text(nObsCase+larger*0.02, ymax*0.8, pCase, color='red', fontsize=ft_size)
        plt.vlines(ymin=0, ymax=ymax, x=nObsCtrl, color='blue', linestyles='--', linewidth=1)
        plt.text(nObsCtrl+larger*0.02, ymax*0.9, pCtrl, color='blue', fontsize=ft_size)
        plt.xlabel('Number of significant tests', fontsize=ft_size, labelpad=5, weight='bold')
        plt.ylabel('Density', fontsize=ft_size, labelpad=5, weight='bold')
        plt.xticks(xticks, fontsize=ft_size)
        plt.yticks(fontsize=ft_size)
        plt.ylim(0, ymax)
        plt.tight_layout()
        plt.close()

        df2 = pd.melt(df, var_name='type')
        fig2 = plt.figure(figsize=(5,4))
        plt.title(plot_title, fontsize=ft_size, loc='left', weight='bold')
        plt.suptitle(subtitle, fontsize=ft_size, y=0.92)
        ax = sns.kdeplot(df2.loc[df2['type']=='case','value'], edgecolor='black', color='#ff8a89', 
                        alpha=.6, fill=True, linewidth=1, label='Case', warn_singular=False)
        ax = sns.kdeplot(df2.loc[df2['type']=='control','value'], edgecolor='black', color='#8b8aff', 
                        alpha=.6, fill=True, linewidth=1, label='Control', warn_singular=False)
        ymax = ax.get_ylim()[-1]
        plt.vlines(ymin=0, ymax=ymax, x=nObsCase, color='red', linestyles='--', linewidth=1)
        plt.text(nObsCase+larger*0.02, ymax*0.7, pCase, color='red', fontsize=ft_size)
        plt.vlines(ymin=0, ymax=ymax, x=nObsCtrl, color='blue', linestyles='--', linewidth=1)
        plt.text(nObsCtrl+larger*0.02, ymax*0.8, pCtrl, color='blue', fontsize=ft_size)
        plt.xlabel('Number of significant tests', fontsize=ft_size, labelpad=5, weight='bold')
        plt.ylabel('Density', fontsize=ft_size, labelpad=5, weight='bold')
        plt.xticks(xticks, fontsize=ft_size)
        plt.yticks(fontsize=ft_size)
        plt.ylim(0, ymax)
        plt.legend(fontsize=7, ncol=2)
        plt.tight_layout()
        plt.close()

        return fig1, fig2        
    
    def _burden_shift_size(self, x, bins):
        if x <= bins[0]:
            return 1
        elif bins[0] < x <= bins[1]:
            return 3
        elif bins[1] < x <= bins[2]:
            return 5
        elif bins[2] < x <= bins[3]:
            return 7
        elif bins[3] < x <= bins[4]:
            return 9
        elif bins[4] < x <= bins[5]:
            return 11
        else:
            return 13

    
    def _change_cre_name(self, x):
        pat = re.search(r'([a-zA-Z]*CRE[a-zA-Z0-9]*)', x).group(1)
        return x.replace(pat, pat.replace('CRE', " ")+'-CRE')
    
    def _match_cat_sets(self, pre, x):
        if len(pre.split("&")) == len(x.replace('is_','').split("_")):
            res = set(pre.split("&"))&set(x.split("_"))
            
            if len(res) == len(pre.split("&")):
                return x
            
    def _create_shiftResPlot_df(self, df, isin_inf=False):
        main_domain = ['Coding', 'PTV', 'Missense', 'Damaging', 'Noncoding', 'Promoter', 'Intron', 'Intergenic', 'UTR', 'LincRNA']
        domain_order = ['Coding (All)','PTV','Missense','Coding w/o PTV','Noncoding (All)','Promoter','Intron','Intergenic','UTR','LincRNA','CRE','Others']
        
        ## modify dataframe as validated form
        df = df.loc[(df['Category_set'] != 'All')&(df['Category_set']!='is_lincRNA_lincRNA')]
        df['Category_set'] = df['Category_set'].str.replace('ptv','PTV')
        df['Category_set'] = df['Category_set'].str.replace('_no_','_w/o_')
        df['Category_set'] = df['Category_set'].str.replace('_wo_','_w/o_')
        df['Domain'] = df['Category_set'].str.split('_').str[1]
        df['Domain'] = df['Domain'].str[0].str.upper() + df['Domain'].str[1:]

        df_case = df.loc[:, ['Category_set','N_cats_case','P_case','Domain']]
        df_case.columns = ['Category_set','N_cats','P','Domain']
        df_case['Phenotype'] = 'Case'
        df_ctrl = df.loc[:, ['Category_set','N_cats_control','P_control','Domain']]
        df_ctrl.columns = ['Category_set','N_cats','P','Domain']
        df_ctrl['Phenotype'] = 'Control'

        df2 = pd.concat([df_case, df_ctrl])#.sort_values(['Category_set','Phenotype'])

        maxs = df2['N_cats'].max()
        mins = df2['N_cats'].min()
        if maxs >= 12:
            num_bins = 6
        elif 6 <= maxs < 12:
            num_bins = 3
        elif 1 < maxs < 6:
            num_bins = 2
        else:
            num_bins = 1
        bins = np.arange(mins, maxs + 1, (maxs - mins) // num_bins)
        df2['Size'] = df2['N_cats'].apply(lambda x: self._burden_shift_size(x, bins))

        df2.loc[~df2.Domain.isin(main_domain), 'Domain'] = 'Others'
        df2["Domain2"] = df2["Domain"]
        df2.loc[df2['Domain2'].isin(["Coding","Noncoding"]), "Domain2"] = df2.loc[df2['Domain2'].isin(["Coding","Noncoding"]), "Domain2"] + " (All)"
        df2.loc[df2['Domain2']=="Damaging", "Domain2"] = "Missense"
        df2.loc[df2['Category_set']=="is_coding_w/o_promoter", "Domain2"] = "Coding w/o PTV"
        df2['Domain_order'] = df2.Domain2.apply(lambda x: domain_order.index(x))

        df2["Category_term"] = df2["Category_set"].str.replace("is_","")
        df2.loc[df2["Category_term"]=='coding_w/o_PTV', "Category_term"] = "coding w/o PTV"
        df2.loc[df2["Category_term"].str.contains("CRE"), "Category_term"] = df2.loc[df2["Category_term"].str.contains("CRE"), "Category_term"].apply(lambda x: self._change_cre_name(x))

        df2.loc[df2['Domain2']=='Others', 'Category_term'] = df2.loc[df2['Domain2']=='Others', 'Category_term'].apply(lambda x: x.replace('_', "\n"))
        df2['Category_term'] = df2["Category_term"].apply(lambda x: " ".join(x.split('_')[1:]) if len(x.split('_'))>2 else x.split('_')[-1])

        df2["Category_term"] = df2["Category_term"].str.replace("CHD8Common", "CHD8 targets")
        df2["Category_term"] = df2["Category_term"].str.replace("FMRPDarnel", "FMRP targets")
        df2["Category_term"] = df2["Category_term"].str.replace("L23", "L2/3")
        df2["Category_term"] = df2["Category_term"].str.replace("L56", "L5/6")
        df2["Category_term"] = df2["Category_term"].str.replace("WillseyUnion", "ASD coexpression")
        df2["Category_term"] = df2["Category_term"].str.replace("ASDTADAFDR03", "ASD risk")
        df2["Category_term"] = df2["Category_term"].str.replace("LOEUF37", "Constrained genes")
        df2["Category_term"] = df2["Category_term"].str.replace("Micro", "Microglia")
        df2["Category_term"] = df2["Category_term"].str.replace("Astro", "Astrocytes")
        df2['Category_term'] = df2['Category_term'].str[0].str.upper() + df2['Category_term'].str[1:]

        df2["new_name"] = df2["Category_term"] + "::" + df2["Domain2"]
        df2["-log10P"] = df2["P"].apply(lambda x: -np.log10(x))
        
        ## extract queried data frame for drawing plot
        all_cat_sets = self.cat_sets.columns.tolist()[1:]
        query_df = pd.DataFrame()
        
        if not self.cat_set_list is None:
            query_cat_sets = []
            for x in self.cat_set_list:
                if len(x.split('&'))>1:
                    query = list(set([x for x in list(map(lambda y: self._match_cat_sets(x, y), all_cat_sets)) if not x is None]))
                    query_cat_sets.extend(query)
                else:
                    query_cat_sets.append('is_'+x)
            query_df = df2.loc[df2["Category_set"].isin(query_cat_sets)]    
        else:
            topN_cats = df2.loc[df2.Phenotype=='Case'].sort_values(["-log10P","N_cats"], ascending=False)["Category_set"].tolist()[:self.n_cat_sets]
            query_df = df2.loc[df2["Category_set"].isin(topN_cats)]

        isin_inf = True if sum(np.isinf(query_df["-log10P"]))>0 else False
        max_lim = df2.loc[~np.isinf(df2["-log10P"]), "-log10P"].max() + 0.5
        query_df.loc[np.isinf(query_df["-log10P"]), "-log10P"] = max_lim
        query_df.sort_values(["Domain_order","Category_set"], ascending=False, inplace=True)   
        
        return (query_df, isin_inf)
    

    def burden_shift(self):
        filt_cats = self.cat_counts.loc[self.cat_counts['Raw_counts'] >= self.c_cutoff, "Category"].tolist()
        filt_cat_sets = self.cat_sets.loc[self.cat_sets['Category'].isin(filt_cats)]
        print_progress("Number of category sets above category counts cutoff: {}".format(len(filt_cat_sets)))
            
        ## Loop through categories specified in the catsets object
        print_progress("Compare burden test and permutation test")
        obsTab = pd.DataFrame()
        output_name = re.sub(r'burden_test\.txt\.gz|burden_test\.txt', '', os.path.basename(self.input_file))
        plot_output = output_name + f'burdenshift_p{self.pval}_cutoff{self.c_cutoff}'
        if self.tag is not None:
            plot_output += f".{self.tag}"
        plot_output += ".dist_plot.pdf"
        pdfsave = PdfPages(os.path.join(self.output_dir_path, plot_output))
        for i in tqdm(range(len(filt_cat_sets.columns))):
            setName = None
            testCats = None
            ## Define the name of the set, and the cateories within it
            if i == 0:
                setName = 'All'
                testCats = filt_cat_sets['Category']
            else:
                setName = filt_cat_sets.columns[i]
                testCats = filt_cat_sets.loc[filt_cat_sets.iloc[:,i]==1, 'Category'].tolist()    
            
            if len(testCats) > 0:
                # Subset true results to only the categories in the set
                burdenResTrim = self.burden_res.filter(self.burden_res['Category'].is_in(testCats))

                # Find case and control counts from this subset of categories (Two-sided)
                nObsCase = len(burdenResTrim.filter((burdenResTrim['P']<=self.pval)&(burdenResTrim['Relative_Risk']>1)))
                nObsCtrl = len(burdenResTrim.filter((burdenResTrim['P']<=self.pval)&(burdenResTrim['Relative_Risk']<1)))
                
                # Subset burden shift results to only the categories in the set
                burdenShitTrim = self.burden_shift_res.select(list(set(self.burden_shift_res.columns)&set(testCats))).to_pandas()
                permCounts = pd.DataFrame(burdenShitTrim.apply(lambda x: self._count_cats(x, self.pval), axis=1).tolist(), columns=['case','control'])
                                
                # Compare the observed counts to the permuted counts to calculate shift p-values
                nPermCase = len(np.where(permCounts['case']>=nObsCase)[0])
                pCase = nPermCase / len(permCounts)
                nPermCtrl = len(np.where(permCounts['control']>=nObsCtrl)[0])
                pCtrl = nPermCtrl / len(permCounts)
                                
                # Plot the results
                fig1, fig2 = self._draw_shiftDistPlot(permCounts, setName, nObsCase, nObsCtrl, pCase, pCtrl)
                pdfsave.savefig(fig1)
                pdfsave.savefig(fig2)
                
                # Add data to output object
                tmp_df = pd.DataFrame([{'Category_set':setName,
                                       "N_cats_case":nObsCase,
                                       'N_cats_control':nObsCtrl,
                                       'P_case':pCase,
                                       'P_control':pCtrl}])
                obsTab = pd.concat([obsTab, tmp_df])
                    
        pdfsave.close()
        
        self._obsTab = obsTab
        
        output_name = re.sub(r'burden_test\.txt\.gz|burden_test\.txt', '', os.path.basename(self.input_file))
        obsFile = output_name + f"burdenshift_p{self.pval}_cutoff{self.c_cutoff}"
        if self.tag is not None:
            obsFile += f".{self.tag}"
        obsFile += ".txt"
        
        obsTab.to_csv(os.path.join(self.output_dir_path, obsFile), sep="\t", index=False)
        
    def draw_shiftResPlot(self):
        output_name = re.sub(r'burden_test\.txt\.gz|burden_test\.txt', '', os.path.basename(self.input_file))
        plot_output = output_name + f'burdenshift_p{self.pval}_cutoff{self.c_cutoff}'
        if self.tag is not None:
            plot_output += f".{self.tag}"
        plot_output += ".result_plot.pdf"
        isin_inf = None
        
        plot_df, isin_inf = self._create_shiftResPlot_df(self._obsTab)
        case_df = plot_df.loc[plot_df.Phenotype=='Case'].reset_index(drop=True)
        ctrl_df = plot_df.loc[plot_df.Phenotype=='Control'].reset_index(drop=True)

        ## Draw plot
        plt.rcParams['font.size'] = self.fontsize
        h = len(plot_df)/2 * 0.7
        fig, ax = plt.subplots(ncols=2, figsize=(14, h),
                               width_ratios=[.5,13.5])

        ## main plot
        ax[1].set_title(self.plot_title, weight='bold', loc='left', pad=5)
        ax[1].scatter(case_df['-log10P'], case_df['new_name'],
                      s=case_df['Size']*20, label='case', color='#ff8a89', edgecolor='black', linewidth=0.5)
        ax[1].scatter(ctrl_df['-log10P'], ctrl_df['new_name'],
                      s=ctrl_df['Size']*20, label='contorl', color='#8b8aff', edgecolor='black', linewidth=0.5)

        y_min, y_max = ax[1].get_ylim()
        ax[1].vlines(ymin=y_min, ymax=y_max, x=-np.log10(0.05), color='red', linestyle='--', linewidth=1.5)

        x_max = ax[1].get_xlim()[-1]
        xticks = np.arange(0,x_max+.5, 0.5)
        xticklabels = [str(x) for x in xticks]
        xticklabels[-2] = 'inf' if isin_inf else xticklabels[-2]
        ax[1].set_xticks(np.arange(0, x_max+.5, 0.5))
        ax[1].set_xticklabels(xticklabels)
        ax[1].set_xlim(0-0.2, round(x_max,1)+0.2)
        ax[1].set_xlabel('P(-log10)')

        ## create and draw grouped y-label
        cat_pos = ax[1].get_yticks()
        cat_labels = [x.get_text() for x in ax[1].get_yticklabels()]
        tmp = pd.DataFrame({'y_pos':cat_pos, 'new_name':cat_labels}, index=range(len(cat_pos)))
        y_axis_df = pd.merge(tmp, plot_df.loc[:,["Category_set","Domain_order","Domain2","new_name"]], on="new_name").drop_duplicates()
        minor_ypos = pd.DataFrame(y_axis_df.groupby("Domain2").mean('y_pos'))["y_pos"].reset_index().rename({"y_pos":"minor_y_pos"}, axis=1)

        ### create vertical line for category domain name
        ax[0].set_yticks(cat_pos)
        ax[0].set_ylim(y_min, y_max)
        for d in y_axis_df["Domain2"].unique():
            y_min_val = y_axis_df.loc[y_axis_df["Domain2"]==d, "y_pos"].min()
            y_max_val = y_axis_df.loc[y_axis_df["Domain2"]==d, "y_pos"].max()
            ax[0].vlines(0, y_min_val-.25, y_max_val+0.25, color='black', linewidth=1.5, ls='-' , clip_on=False)
            
        ### add domain name (grouping y-label)
        for y_pos, y_label in zip(minor_ypos["minor_y_pos"], minor_ypos['Domain2']):
            ax[0].text(-.015, y_pos, y_label, color='black', ha="right", va='center')
        ax[0].axis('off')

        ## set yaxis of main plot
        yticklabels = [text.split("::")[0] for text in cat_labels]
        ax[1].yaxis.set_major_locator(mticker.FixedLocator(cat_pos))
        ax[1].set_yticklabels(yticklabels)
        ax[1].set_ylim(y_min, y_max)

        ## create multiple legends
        ### legend1 - phenotype
        pheno_lds = [Line2D([0],[0], markerfacecolor='#ff8a89', marker='o', markersize=self.fontsize*1.15, markeredgewidth=0.5, linewidth=0, markeredgecolor='black', label='Case'),
                     Line2D([0],[0], markerfacecolor='#8b8aff', marker='o', markersize=self.fontsize*1.15, markeredgewidth=0.5, linewidth=0, markeredgecolor='black', label='Control')]
        legend1 = ax[1].legend(handles=pheno_lds, loc='center left', bbox_to_anchor=(1,0.25), labelspacing=.7, title='Phenotype', frameon=False, borderaxespad=0.)
        legend1._legend_box.align = 'left'

        ### legend2 - size
        size_lds = []
        sizes = [1, 3, 5, 7, 9, 11, 13]
        size_labels = ['0','<10','<50','<100','<150','<200',"≥200"]
        for s, sl in zip(sizes, size_labels):
            tmp_lds = Line2D([0],[0], markerfacecolor='white', marker='o', markersize=s, markeredgewidth=0.5, markeredgecolor='black', linewidth=0, label=sl)
            size_lds.append(tmp_lds)
        legend2 = ax[1].legend(handles=size_lds, loc='center left', bbox_to_anchor=(1,0.6), labelspacing=.7, title='Number of\nsignificant\ncategories', frameon=False, borderaxespad=0.)
        legend2._legend_box.align = 'left'

        ax[1].add_artist(legend1)
        ax[1].add_artist(legend2)
        ax[1].legend(handles=[], loc='center left', bbox_to_anchor=(1.1,0.5), frameon=False)
        
        ## set ticks params
        ax[1].spines['top'].set_visible(False)
        ax[1].spines['right'].set_visible(False)
        ax[1].spines['left'].set_linewidth(1.5)
        ax[1].spines['bottom'].set_linewidth(1.5)
        ax[1].tick_params(width=1.5, size=7)

        plt.tight_layout()
        plt.savefig(os.path.join(self.output_dir_path, plot_output), bbox_inches='tight')    

    def run(self):
        self.burden_shift()
        self.draw_shiftResPlot()
        print_progress("Done")


"""
Command line generator for Variant Effect Predictor (VEP)
"""
from cwas.utils.check import check_is_file
from cwas.utils.check import check_is_dir


class VepCmdGenerator:
    def __init__(self, vep_path: str,
                 vep_cache_path: str, vep_conservation_path: str,
                 vep_loftee_path: str, vep_human_ancestor_fa_path: str,
                 vep_gerp_bw_path: str, vep_mis_db_path: str, 
                 vep_mis_info_key: str, input_vcf_path: str, 
                 num_proc: str) -> None:
        self._vep_path = vep_path
        self._vep_cache_path = vep_cache_path
        self._vep_conservation_path = vep_conservation_path
        self._vep_loftee_path = vep_loftee_path
        self._vep_human_ancestor_fa_path = vep_human_ancestor_fa_path
        self._vep_gerp_bw_path = vep_gerp_bw_path
        self._vep_mis_db_path = vep_mis_db_path
        self._vep_mis_info_key = vep_mis_info_key
        self._input_vcf_path = input_vcf_path
        self._check_validity()
        self._output_vcf_path = input_vcf_path.replace(".vcf", ".vep.vcf")
        self._num_proc = num_proc

    @staticmethod
    def _check_path(path: str, message: str, is_dir: bool = False):
        try:
             if is_dir:
                 check_is_dir(path)
             else:
                 check_is_file(path)
        except ValueError:
            raise ValueError(f"{message}: {path}")
        except Exception:
            raise

    def _check_validity(self):
         self._check_path(self._vep_path, "Invalid VEP path")
         self._check_path(self._vep_conservation_path, "Invalid VEP resource path (conservation file)")
         self._check_path(self._vep_loftee_path, "Invalid VEP resource path (loftee directory)", is_dir=True)
         self._check_path(self._vep_human_ancestor_fa_path, "Invalid VEP resource path (human ancestor fasta file)")
         self._check_path(self._vep_gerp_bw_path, "Invalid VEP resource path (gerp bigwig file)")
         self._check_path(self._vep_mis_db_path, "Invalid VEP resource path (missense database file)")
         self._check_path(self._input_vcf_path, "Invalid input VCF path")
         self._check_path(self._vep_cache_path, "Invalid VEP cache directory path", is_dir=True)

    @property
    def vep_path(self) -> str:
        return self._vep_path

    @property
    def vep_conservation_path(self) -> str:
        return self._vep_conservation_path

    @property
    def vep_loftee_path(self) -> str:
        return self._vep_loftee_path

    @property
    def vep_human_ancestor_fa_path(self) -> str:
        return self._vep_human_ancestor_fa_path

    @property
    def vep_gerp_bw_path(self) -> str:
        return self._vep_gerp_bw_path

    @property
    def vep_mis_db_path(self) -> str:
        return self._vep_mis_db_path

    @property
    def vep_mis_info_key(self) -> str:
        return self._vep_mis_info_key

    @property
    def vep_cache_path(self) -> str:
        return self._vep_cache_path

    @property
    def input_vcf_path(self) -> str:
        return self._input_vcf_path

    @property
    def output_vcf_path(self) -> str:
        return self._output_vcf_path

    @output_vcf_path.setter
    def output_vcf_path(self, arg: str):
        self._output_vcf_path = arg

    @property
    def num_proc(self) -> str:
        return self._num_proc

    @property
    def cmd_str(self) -> str:
        return " ".join(self.cmd)

    @property
    def cmd(self) -> list:
        result = [
            self.vep_path,
            "-i",
            self.input_vcf_path,
            "-o",
            self.output_vcf_path,
        ]
        result += self.cmd_option_basic
        result += self.cmd_option_pick_one_gene_isoform
        result += self.cmd_option_pick_nearest_gene
        return result

    @property
    def cmd_option_basic(self) -> list:
        """Return basic options (no plugins) of VEP"""
        return [
            "--assembly",
            "GRCh38",
            "--offline",
            "--cache",
            "--dir_cache",
            self.vep_cache_path,
            "--force_overwrite",
            "--format",
            "vcf",
            "--vcf",
            "--no_stats",
            "--plugin",
            ','.join(['LoF',
                      'conservation_file:' + self.vep_conservation_path,
                      'loftee_path:' + self.vep_loftee_path,
                      'human_ancestor_fa:' + self.vep_human_ancestor_fa_path,
                      'gerp_bigwig:' + self.vep_gerp_bw_path]),
            "--dir_plugins",
            self.vep_loftee_path,
            "--custom",
            ",".join([self._vep_mis_db_path, "MisDb", 'vcf', "exact", "0", self.vep_mis_info_key]),
        ]

    @property
    def cmd_option_pick_one_gene_isoform(self) -> list:
        """Return options in order to pick a gene isoform
        with most severe consequence"""
        return [
            "--per_gene",
            "--pick",
            "--pick_order",
            "rank,canonical,appris,tsl,biotype,ccds,length",
        ]

    @property
    def cmd_option_pick_nearest_gene(self) -> list:
        """Return options in order to pick the nearest gene"""
        return ["--distance", "2000", "--nearest", "symbol", "--symbol"]


import argparse
from abc import abstractmethod
from pathlib import Path
from typing import Optional
import polars as pl
import numpy as np
import pandas as pd
import re

from cwas.core.categorization.category import Category
from cwas.core.common import cmp_two_arr
from cwas.runnable import Runnable
from cwas.utils.check import check_is_file
from cwas.utils.check import check_is_dir
from cwas.utils.log import print_arg, print_progress



class BurdenTest(Runnable):
    def __init__(self, args: Optional[argparse.Namespace] = None):
        super().__init__(args)
        self._sample_info = None
        self._adj_factor = None
        self._categorization_result = None
        self._result = None
        self._phenotypes = None
        self._case_variant_cnt = None
        self._ctrl_variant_cnt = None
        self._case_carrier_cnt = None
        self._ctrl_carrier_cnt = None
        self._count_thres = None

    @staticmethod
    def _print_args(args: argparse.Namespace):
        print_arg("Categorized file", args.cat_path)
        print_arg("Output directory", args.output_dir_path)
        print_arg("Sample information file", args.sample_info_path)
        print_arg("Adjustment factor list", args.adj_factor_path)
        print_arg("If the number of carrier is used for burden test or not", args.use_n_carrier)

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.cat_path)
        check_is_dir(args.output_dir_path)
        check_is_file(args.sample_info_path)
        if args.adj_factor_path is not None:
            check_is_file(args.adj_factor_path)

    @property
    def cat_path(self) -> Path:
        return self.args.cat_path.resolve()

    @property
    def output_dir_path(self) -> Path:
        return self.args.output_dir_path.resolve()

    @property
    def sample_info_path(self) -> Path:
        return self.args.sample_info_path.resolve()

    @property
    def adj_factor_path(self) -> Optional[Path]:
        return (
            self.args.adj_factor_path.resolve()
            if self.args.adj_factor_path
            else None
        )

    @property
    def result_path(self) -> Path:
        f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'burden_test.txt', self.cat_path.name)
        return Path(
            f"{self.output_dir_path}/"
            f"{f_name}"
        )

    @property
    def counts_path(self) -> Path:
        f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'category_counts.txt', self.cat_path.name)
        return Path(
            f"{self.output_dir_path}/"
            f"{f_name}"
        )

    @property
    def cat_info_path(self) -> Path:
        f_name = re.sub(r'categorization_result\.txt\.gz|categorization_result\.txt', 'category_info.txt', self.cat_path.name)
        return Path(
            f"{self.output_dir_path}/"
            f"{f_name}"
        )

    @property
    def sample_info(self) -> pd.DataFrame:
        if self._sample_info is None:
            self._sample_info = pd.read_table(
                self.sample_info_path, index_col="SAMPLE", dtype={"SAMPLE": str}, sep="\t"
            )
        return self._sample_info

    @property
    def adj_factor(self) -> pd.DataFrame:
        if self._adj_factor is None and self.adj_factor_path:
            self._adj_factor = pd.read_table(
                self.adj_factor_path, index_col="SAMPLE", dtype={"SAMPLE": str}, sep="\t"
            )
        return self._adj_factor

    @property
    def use_n_carrier(self) -> bool:
        return self.args.use_n_carrier

    @property
    def tag(self) -> str:
        return self.args.tag

    @property
    def eff_test(self) -> int:
        return self.args.eff_test

    @property
    def marker_size(self) -> float:
        return self.args.marker_size

    @property
    def font_size(self) -> float:
        return self.args.font_size

    @property
    def plot_size(self) -> float:
        return self.args.plot_size
    
    @property
    def plot_title(self) -> float:
        return self.args.plot_title

    @property
    def categorization_result(self) -> pd.DataFrame:
        if self._categorization_result is None:
            print_progress("Load the categorization result")
            self._categorization_result = pl.read_csv(
                self.cat_path, separator="\t", dtypes={"SAMPLE": str}
            )
            self._categorization_result = self._categorization_result.to_pandas().set_index("SAMPLE")
            self.save_counts_table(form = 'raw')
            if self.adj_factor is not None:
                self._adjust_categorization_result()
        return self._categorization_result

    def _adjust_categorization_result(self):
        if not _contain_same_index(
            self._categorization_result, self.adj_factor
        ):
            raise ValueError(
                "The sample IDs from the adjustment factor list are "
                "not the same with the sample IDs "
                "from the categorization result."
            )
        adj_factors = [
            self.adj_factor.to_dict()["AdjustFactor"][sample_id]
            for sample_id in self._categorization_result.index.values
        ]
        self._categorization_result = self._categorization_result.multiply(
            adj_factors, axis="index"
        )

    @property
    def phenotypes(self) -> np.ndarray:
        if self._phenotypes is None:
            self._phenotypes = np.array(self.sample_info.loc[self.categorization_result.index, 'PHENOTYPE'])
            
        return self._phenotypes

    @property
    def case_cnt(self) -> int:
        return (self.phenotypes == "case").sum()

    @property
    def ctrl_cnt(self) -> int:
        return (self.phenotypes == "ctrl").sum()

    @property
    def case_variant_cnt(self) -> np.ndarray:
        if self._case_variant_cnt is None:
            self._case_variant_cnt = self.categorization_result.values[
                self.phenotypes == "case", :
            ].sum(axis=0)
        return self._case_variant_cnt

    @property
    def ctrl_variant_cnt(self) -> np.ndarray:
        if self._ctrl_variant_cnt is None:
            self._ctrl_variant_cnt = self.categorization_result.values[
                self.phenotypes == "ctrl", :
            ].sum(axis=0)
        return self._ctrl_variant_cnt

    @property
    def case_carrier_cnt(self) -> np.ndarray:
        if self._case_carrier_cnt is None:
            var_counts = self.categorization_result.values
            is_carrier = np.where(var_counts > 0, 1, 0)
            self._case_carrier_cnt = is_carrier[
                self.phenotypes == "case", :
            ].sum(axis=0)
        return self._case_carrier_cnt

    @property
    def ctrl_carrier_cnt(self) -> np.ndarray:
        if self._ctrl_carrier_cnt is None:
            var_counts = self.categorization_result.values
            is_carrier = np.where(var_counts > 0, 1, 0)
            self._ctrl_carrier_cnt = is_carrier[
                self.phenotypes == "ctrl", :
            ].sum(axis=0)
        return self._ctrl_carrier_cnt

    @property
    def category_table(self) -> pd.DataFrame:
        categories = [
            Category.from_str(category_str).to_dict()
            for category_str in self.categorization_result.columns.values
        ]
        return pd.DataFrame(
            categories, index=self.categorization_result.columns.values
        )

    def run(self):
        if not _contain_same_index(
            self.categorization_result, self.sample_info
        ):
            raise ValueError(
                "The sample IDs from the sample information are "
                "not the same with the sample IDs "
                "from the categorization result."
            )
        if self.use_n_carrier:
            self.count_carrier_for_each_category()
            self.calculate_relative_risk_with_n_carrier()
        else:
            self.count_variant_for_each_category()
            self.calculate_relative_risk()        
        self.run_burden_test()
        self.concat_category_info()
        self.save_result()
        self.save_counts_table(form = 'adj')
        self.save_category_info()
        self.update_env()

    def count_variant_for_each_category(self):
        print_progress("Count the number of variants for each category")
        variant_cnt_arr = np.concatenate(
            [
                self.case_variant_cnt[:, np.newaxis],
                self.ctrl_variant_cnt[:, np.newaxis],
            ],
            axis=1,
        )
        self._result = pd.DataFrame(
            variant_cnt_arr,
            index=self.categorization_result.columns.values,
            columns=["Case_DNV_Count", "Ctrl_DNV_Count"],
        )

    def count_carrier_for_each_category(self):
        print_progress("Count the number of carriers in each category")
        carrier_cnt_arr = np.concatenate(
            [
                self.case_carrier_cnt[:, np.newaxis],
                self.ctrl_carrier_cnt[:, np.newaxis],
            ],
            axis=1,
        )
        self._result = pd.DataFrame(
            carrier_cnt_arr,
            index=self.categorization_result.columns.values,
            columns=["Case_Carrier_Count", "Ctrl_Carrier_Count"],
        )

    def calculate_relative_risk(self):
        print_progress("Calculate relative risks for each category")
        np.seterr(divide='ignore')
        normalized_case_variant_cnt = self.case_variant_cnt / self.case_cnt
        normalized_ctrl_variant_cnt = self.ctrl_variant_cnt / self.ctrl_cnt
        self._result["Relative_Risk"] = (
            normalized_case_variant_cnt / normalized_ctrl_variant_cnt
        )

    def calculate_relative_risk_with_n_carrier(self):
        print_progress("Calculate relative risks for each category")
        case_carrier_rate = self.case_carrier_cnt / self.case_cnt
        ctrl_carrier_rate = self.ctrl_carrier_cnt / self.ctrl_cnt
        self._result["Relative_Risk"] = (
            case_carrier_rate / ctrl_carrier_rate
        )

    @abstractmethod
    def run_burden_test(self):
        raise RuntimeError(
            "This method cannot be called via the instance of BurdenTest."
        )

    def concat_category_info(self):
        self._result = pd.concat([self.category_table, self._result], axis=1)
        self._result.index.name = "Category"

    def save_result(self):
        print_progress(f"Save the result to the file {self.result_path}")
        self._result.to_csv(self.result_path, sep="\t")

    def update_env(self):
        self.set_env("BURDEN_TEST_RESULT", self.result_path)
        self.save_env()
        
    def save_counts_table(self, form: str):
        if form == 'raw':
            if self.use_n_carrier:
                self._raw_counts = pd.DataFrame({'Raw_counts': (self.categorization_result > 0).sum(axis=0)},
                                                index=self.categorization_result.columns)
                self._raw_counts.index.name = 'Category'

            else:
                self._raw_counts = pd.DataFrame({'Raw_counts': self.categorization_result.sum(axis=0)},
                                                index= self.categorization_result.sum(axis=0).index)
                self._raw_counts.index.name = 'Category'
        elif form =='adj':
            if self.use_n_carrier:
                self._adj_counts = pd.DataFrame({'Adj_counts': self._result['Case_Carrier_Count'] + self._result['Ctrl_Carrier_Count']})
            else:
                self._adj_counts = pd.DataFrame({'Adj_counts': self._result['Case_DNV_Count'] + self._result['Ctrl_DNV_Count']})
            self._counts_table = pd.merge(self._raw_counts, self._adj_counts, on='Category')
            print_progress(f"Save the category counts to the file {self.counts_path}")
            self._counts_table.to_csv(self.counts_path, sep="\t")
    
    def save_category_info(self):
        cat_set = self._result.loc[:, ['variant_type', 'gene_set', 'functional_score', 'gencode', 'functional_annotation']]
        cat_set = apply_region_mapping(cat_set)
        print_progress(f"Save the category information to the file {self.cat_info_path}")
        cat_set.to_csv(self.cat_info_path, sep="\t")



def _contain_same_index(table1: pd.DataFrame, table2: pd.DataFrame) -> bool:
    return cmp_two_arr(table1.index.values, table2.index.values)

def apply_region_mapping(df: pd.DataFrame):
    coding_region = ['CodingRegion', 'MissenseRegion', 'SilentRegion', 'PTVRegion',
                     'InFrameRegion', 'DamagingMissenseRegion', 'FrameshiftRegion']
    noncoding_region = ['NoncodingRegion', 'IntronRegion', 'lincRnaRegion',
                        'IntergenicRegion', 'OtherTranscriptRegion', 'PromoterRegion',
                        'UTRsRegion', 'SpliceSiteNoncanonRegion']
    
    region_mapping = {
        'is_coding': lambda x: x['gencode'].isin(coding_region).astype(int),
        'is_coding_no_ptv': lambda x: x['gencode'].isin(set(coding_region) - set(['CodingRegion', 'PTVRegion', 'FrameshiftRegion'])).astype(int),
        'is_PTV': lambda x: (x['gencode'] == 'PTVRegion').astype(int),
        'is_missense': lambda x: (x['gencode'] == 'MissenseRegion').astype(int),
        'is_damaging_missense': lambda x: (x['gencode'] == 'DamagingMissenseRegion').astype(int),
        'is_noncoding': lambda x: x['gencode'].isin(noncoding_region).astype(int),
        'is_noncoding_wo_promoter': lambda x: x['gencode'].isin(set(noncoding_region) - set(['NoncodingRegion', 'PromoterRegion'])).astype(int),
        'is_promoter': lambda x: (x['gencode'] == 'PromoterRegion').astype(int),
        'is_intron': lambda x: (x['gencode'] == 'IntronRegion').astype(int),
        'is_intergenic': lambda x: (x['gencode'] == 'IntergenicRegion').astype(int),
        'is_UTR': lambda x: (x['gencode'] == 'UTRsRegion').astype(int),
        'is_lincRNA': lambda x: ((x['gene_set'] == 'lincRNA') | (x['gencode'] == 'lincRnaRegion')).astype(int)
    }

    for col, condition in region_mapping.items():
        df[col] = condition(df)
    
    return df


import argparse
import shutil
from pathlib import Path
import os

import cwas.core.configuration.create as create
import cwas.utils.log as log
import cwas.utils.check as check
from cwas.runnable import Runnable


class Configuration(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.workspace = self._get_workspace()
        self.user_config = self.workspace / "configuration.txt"
        self.data_dir_symlink = self.workspace / "annotation-data"
        self.gene_matrix_symlink = self.workspace / "gene_matrix.txt"
        self.bed_key_list_symlink = self.workspace / "annotation_keys.yaml"
        self.category_domain_list = self.workspace / "category_domain.yaml"
        self.redundant_category_table = (
            self.workspace / "redundant_category.txt"
        )

    def _get_workspace(self):
        workspace = self.get_env("CWAS_WORKSPACE")
        if workspace is None:
            raise RuntimeError(
                "${CWAS_WORKSPACE} does not set. Run the Start step first."
            )
        return Path(workspace)

    @staticmethod
    def _print_args(args: argparse.Namespace):
        log.print_arg(
            "Force to overwrite configuration",
            "Y" if args.force_overwrite else "N",
        )

    @property
    def force_overwrite(self):
        return bool(self.args.force_overwrite)

    def run(self):
        self._set_config_to_attr()
        self._check_attr_from_user_config()
        self._create_data_dir_symlink()
        self._create_gene_matrix_symlink()
        self._create_bed_key_list_symlink()
        self._create_category_info()
        self._set_env()

    def _set_config_to_attr(self):
        user_config = self._load_configuration()
        self.data_dir = Path(user_config.get("ANNOTATION_DATA_DIR"))
        self.gene_matrix = self.data_dir.joinpath(user_config.get("GENE_MATRIX"))
        self.vep = Path(user_config.get("VEP"))
        self.vep_cache_dir = Path(user_config.get("VEP_CACHE_DIR"))
        self.vep_conservation = self.vep_cache_dir.joinpath(user_config.get("VEP_CONSERVATION_FILE"))
        self.vep_loftee = self.vep_cache_dir.joinpath(user_config.get("VEP_LOFTEE"))
        self.vep_human_ancestor_fa = self.vep_cache_dir.joinpath(user_config.get("VEP_HUMAN_ANCESTOR_FA"))
        self.vep_gerp_bw = self.vep_cache_dir.joinpath(user_config.get("VEP_GERP_BIGWIG"))
        self.vep_mis_db = self.vep_cache_dir.joinpath(user_config.get("VEP_MIS_DB"))
        self.vep_mis_info_key = user_config.get("VEP_MIS_INFO_KEY")
        self.vep_mis_thres = Path(user_config.get("VEP_MIS_THRES"))

        annot_key_conf = user_config.get("ANNOTATION_KEY_CONFIG")
        self.annot_key_conf = (
            None if not annot_key_conf else self.data_dir.joinpath(annot_key_conf)
        )

    def _load_configuration(self) -> dict:
        user_config_dict = {}

        with self.user_config.open() as user_config_file:
            for line in user_config_file:
                k, v = line.strip().split("=")
                # The two 'CONFIG' file paths are optional.
                if not k.endswith("CONFIG"):
                    self._check_config_value(k, v)
                user_config_dict[k] = v

        return user_config_dict

    def _check_config_value(self, key: str, value: str):
        if not value:
            raise ValueError(f"The value for '{key}' is empty.")

    def _check_attr_from_user_config(self):
        check.check_is_file(self.user_config)
        check.check_is_dir(self.data_dir)
        check.check_is_dir(self.vep_cache_dir)        
        check.check_is_file(self.vep_conservation)
        check.check_is_dir(self.vep_loftee)
        check.check_is_file(self.vep_human_ancestor_fa)
        check.check_is_file(self.vep_gerp_bw)
        check.check_is_file(self.vep_mis_db)
        check.check_is_file(self.gene_matrix)
        if self.annot_key_conf is not None:
            check.check_is_file(self.annot_key_conf)
        

        if self.vep is not None and shutil.which(self.vep) is None:
            raise ValueError(f'"{self.vep} is not an executable."')

    def _create_data_dir_symlink(self):
        data_dir = getattr(self, "data_dir")
        data_dir_symlink = getattr(self, "data_dir_symlink")
        log.print_progress(
            f"Create a symlink to your annotation data "
            f'directory "{data_dir_symlink}"'
        )
        try:
            data_dir_symlink.symlink_to(data_dir, target_is_directory=True)
        except FileExistsError:
            if self.force_overwrite:
                log.print_warn(
                    f'"{data_dir_symlink}" already exists, removing it and creating the symbolic link again.'
                )
                temp_link = Path(str(data_dir_symlink) + ".new")
                os.unlink(data_dir_symlink)
                os.symlink(data_dir, temp_link)
                os.rename(temp_link, data_dir_symlink)
            else:
                log.print_warn(
                    f'"{data_dir_symlink}" already exists so skip '
                    f"creating the symbolic link"
                )

    def _create_gene_matrix_symlink(self):
        gene_matrix = getattr(self, "gene_matrix")
        gene_matrix_symlink = getattr(self, "gene_matrix_symlink")
        log.print_progress(
            f"Create a symlink to your gene matrix " f'"{gene_matrix_symlink}"'
        )

        try:
            gene_matrix_symlink.symlink_to(gene_matrix)
        except FileExistsError:
            if self.force_overwrite:
                log.print_warn(
                    f'"{gene_matrix_symlink}" already exists, removing it and creating the symbolic link again.'
                )
                temp_link = Path(str(gene_matrix_symlink) + ".new")
                os.remove(gene_matrix_symlink)
                os.symlink(gene_matrix, temp_link)
                os.rename(temp_link, gene_matrix_symlink)
            else:
                log.print_warn(
                    f'"{gene_matrix_symlink}" already exists so skip '
                    f"creating the symbolic link"
                )
            
    def _create_bed_key_list_symlink(self):
        annot_key_conf = getattr(self, "annot_key_conf")
        bed_key_list_symlink = getattr(self, "bed_key_list_symlink")
        log.print_progress(
            f"Create a symlink to your annotation key list " f'"{bed_key_list_symlink}"'
        )

        try:
            bed_key_list_symlink.symlink_to(annot_key_conf)
        except FileExistsError:

            if self.force_overwrite:
                log.print_warn(
                    f'"{bed_key_list_symlink}" already exists, removing it and creating the symbolic link again.'
                )
                temp_link = Path(str(bed_key_list_symlink) + ".new")
                os.remove(bed_key_list_symlink)
                os.symlink(annot_key_conf, temp_link)
                os.rename(temp_link, bed_key_list_symlink)
            else:
                log.print_warn(
                    f'"{bed_key_list_symlink}" already exists so skip '
                    f"creating the symbolic link"
                )

    def _create_category_info(self):
        """ Create a list of category domains and a redundant category table"""
        domain_list = getattr(self, "category_domain_list")
        annot_key_conf = getattr(self, "annot_key_conf")
        gene_matrix = getattr(self, "gene_matrix")
        redundant_category_table = getattr(self, "redundant_category_table")

        log.print_progress(
            f"Create a CWAS category domain list " f'"{domain_list}"'
        )
        create.create_category_domain_list(
            domain_list, annot_key_conf, gene_matrix
        )
        log.print_progress(
            f"Create a redundant category table "
            f'"{redundant_category_table}"'
        )
        create.create_redundant_category_table(redundant_category_table)

    def _set_env(self):
        log.print_progress("Set CWAS environment variables")
        self.set_env("VEP", getattr(self, "vep"))
        self.set_env("VEP_CACHE_DIR", getattr(self, "vep_cache_dir"))
        self.set_env("VEP_CONSERVATION_FILE", getattr(self, "vep_conservation"))
        self.set_env("VEP_LOFTEE", getattr(self, "vep_loftee"))
        self.set_env("VEP_HUMAN_ANCESTOR_FA", getattr(self, "vep_human_ancestor_fa"))
        self.set_env("VEP_GERP_BIGWIG", getattr(self, "vep_gerp_bw"))
        self.set_env("VEP_MIS_DB", getattr(self, "vep_mis_db"))
        self.set_env("VEP_MIS_INFO_KEY", getattr(self, "vep_mis_info_key"))
        self.set_env("VEP_MIS_THRES", getattr(self, "vep_mis_thres"))
        self.set_env("ANNOTATION_DATA", self.data_dir_symlink)
        self.set_env("GENE_MATRIX", self.gene_matrix_symlink)
        self.set_env("ANNOTATION_BED_KEY", self.bed_key_list_symlink)
        self.set_env("CATEGORY_DOMAIN", self.category_domain_list)
        self.set_env("REDUNDANT_CATEGORY", self.redundant_category_table)
        self.save_env()



"""
Command line tool for Category-wide association study (CWAS)
"""
import sys

import cwas.factory
from cwas.utils.log import print_log


def main():
    # TODO: Add conditional statements or exception handling. This is very
    #  error-prone.
    print_log("LOG", "Category-Wide Association Study (CWAS)")
    cwas_factory = cwas.factory.create(sys.argv[1])
    cwas_args = cwas_factory.argparser().parse_args(sys.argv[2:])
    cwas_obj = cwas_factory.runnable
    print_log("LOG", f"Current step: {cwas_obj.__name__}")
    cwas_inst = cwas_obj(cwas_args)
    cwas_inst.run()



def get_idx_dict(list_: list) -> dict:
    return {item: i for i, item in enumerate(list_)}


def extract_sublist_by_int(list_: list, n: int) -> list:
    """Get a sublist from the input list by using the input integer"""
    i = 0
    result = []

    while n != 0:
        if n % 2 == 1:
            result.append(list_[i])
        n >>= 1
        i += 1

    return result



"""
Factory method for Runnable
Ref: https://realpython.com/python-import/
"""
import argparse
import importlib

import cwas.argparser
from cwas.runnable import Runnable


class CWASFactory:
    def __init__(
        self, argparser: argparse.ArgumentParser, runnable: Runnable
    ) -> None:
        self._argparser = argparser
        self._runnable = runnable

    @property
    def argparser(self) -> argparse.ArgumentParser:
        return self._argparser

    @property
    def runnable(self) -> Runnable:
        return self._runnable


def create(step_name: str) -> CWASFactory:
    try:
        runnable_module = importlib.import_module(f"cwas.{step_name}")
        runnable = getattr(runnable_module, make_class_name(step_name))
        argparser = getattr(cwas.argparser, step_name)
    except (ImportError, AttributeError):
        raise ValueError(f'CWAS does not support "{step_name}".')

    return CWASFactory(argparser, runnable)


def make_class_name(module_name: str) -> str:
    class_name = module_name.title()
    class_name = "".join(class_name.split("_"))
    return class_name



"""
A container class that contains iterator protocol to read a BED file
"""
from pathlib import Path

import pysam
from cwas.utils.log import print_warn


class BedReader:
    def __init__(self, bed_path: Path):
        self.bed_path = bed_path
        self.contig = None
        self.start = None
        self.stop = None

        if not self.bed_path.exists():
            raise OSError(
                f"Failed to initialize a BedReader instance. "
                f"'{self.bed_path}' does not exists."
            )

    def __iter__(self):
        try:
            with pysam.TabixFile(str(self.bed_path)) as bed_file:
                for fields in bed_file.fetch(
                    self.contig, self.start, self.stop, parser=pysam.asTuple()
                ):
                    contig, start, stop, *others = tuple(fields)
                    yield (contig, int(start), int(stop), *others)
        except ValueError:
            print_warn(
                f"The contig '{self.contig}' does not exist "
                f"in the bed file '{self.bed_path}'."
            )

    def set_contig(self, contig: str):
        self.contig = contig



import argparse, os, sys, pickle
from pathlib import Path
import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Optional
import re

from cwas.utils.log import print_progress, print_arg, print_log
from cwas.runnable import Runnable
from scipy.stats import norm
from cwas.utils.check import check_is_file, check_is_dir
from scipy.stats import binomtest

class EffectiveNumTest(Runnable):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self._intersection_matrix = None
        self._correlation_matrix = None
        self._category_set_path = None
        self._category_set = None
        self._num_eig = None
        self.eff_num_test_value = None
        self._replace_term = None
        self._binom_p = None
        self._sample_info = None
        
    @staticmethod
    def _print_args(args: argparse.Namespace):
        print_arg("Input file", args.input_path)
        print_arg("Input format", args.input_format)
        print_arg("Output directory", args.output_dir_path)
        if args.sample_info_path:
            print_arg("Sample information file", args.sample_info_path)
        if args.category_count_file:
            print_arg("Using variant (or sample) counts file: ", args.category_count_file)
        if args.tag:
            print_arg("Output tag (prefix of output files)", args.tag)
        if args.category_set_path :
            print_arg("Category set file", args.category_set_path)

    @staticmethod
    def _check_args_validity(args: argparse.Namespace):
        check_is_file(args.input_path)
        check_is_dir(args.output_dir_path)
        if args.sample_info_path:
            check_is_file(args.sample_info_path)
        if args.category_count_file:
            check_is_file(args.category_count_file)
        if args.category_set_path :
            check_is_file(args.category_set_path)

    @property
    def input_path(self) -> Path:
        return self.args.input_path.resolve()

    @property
    def output_dir_path(self) -> Path:
        return(self.args.output_dir_path.resolve())

    @property
    def input_format(self) -> str:
        return self.args.input_format

    @property
    def category_count(self):
        category_count_ = pd.read_table(self.args.category_count_file, sep="\t")
        return category_count_
    
    @property
    def intersection_matrix(self) -> pd.DataFrame:
        if self._intersection_matrix is None and self.input_format == 'inter':
            with self.input_path.open('rb') as f:
                self._intersection_matrix = pickle.load(f)
            self._intersection_matrix.index = list(map(str, self._intersection_matrix.index.tolist()))
            self._intersection_matrix.columns = list(map(str, self._intersection_matrix.columns.tolist()))
        return self._intersection_matrix

    @property
    def correlation_matrix(self) -> pd.DataFrame:
        if self._correlation_matrix is None and self.input_format == 'corr':
            with self.input_path.open('rb') as f:
                self._correlation_matrix = pickle.load(f)
            self._correlation_matrix.index = list(map(str, self._correlation_matrix.index.tolist()))
            self._correlation_matrix.columns = list(map(str, self._correlation_matrix.columns.tolist()))
        return self._correlation_matrix

    @property
    def sample_info_path(self) -> Optional[Path]:
        return self.args.sample_info_path.resolve()

    @property
    def sample_info(self) -> pd.DataFrame:
        if self._sample_info is None:
            self._sample_info = pd.read_table(
                self.sample_info_path, index_col="SAMPLE", dtype={"SAMPLE": str}, sep="\t"
            )
        return self._sample_info

    @property
    def binom_p(self) -> float:
        if self._binom_p is None:
            self._binom_p = (self.sample_info["PHENOTYPE"] == "case").sum() / np.isin(self.sample_info["PHENOTYPE"], ["case", "ctrl"]).sum()
        return self._binom_p

    @property
    def count_thres(self) -> int:
        if self.args.count_thres is None:
            m = 1
            while True:
                p_value = binomtest(m-1, m, self.binom_p, alternative='greater').pvalue
                if p_value < 0.05:
                    return m
                m += 1
        else:
            return self.args.count_thres

    @property
    def num_eig(self) -> int:
        if self._num_eig is None:
            self._num_eig = self.args.num_eig
            return self._num_eig

    @property
    def category_set_path(self) -> Optional[Path]:
        return (
            self.args.category_set_path.resolve()
            if self.args.category_set_path
            else None
        )
    @property
    def category_set(self) -> pd.DataFrame:
        if self._category_set is None and self.category_set_path:
            self._category_set = pd.read_csv(self.category_set_path, sep='\t')
        return self._category_set

    @property
    def eff_num_test(self) -> bool:
        return self.args.eff_num_test

    @property
    def tag(self) -> str:
        return self.args.tag

    @property
    def replace_term(self) -> str:
        if self._replace_term is None:
            if 'correlation_matrix.pkl' in str(self.input_path):
                self._replace_term = r'\.correlation_matrix\.pkl\.gz|\.correlation_matrix\.pkl'
            elif 'intersection_matrix.pkl' in str(self.input_path):
                self._replace_term = r'\.intersection_matrix\.pkl\.gz|\.intersection_matrix\.pkl'
            else:
                self._replace_term = r'\.txt\.gz|\.txt'
        return self._replace_term

    @property
    def corr_mat_path(self) -> Path:
        if self.tag is None:
            save_name = '.correlation_matrix.pickle'
        else:
            save_name = '.'.join(['.correlation_matrix', self.tag, 'pickle'])
        f_name = re.sub(self.replace_term, save_name, self.input_path.name)
        return Path(
            f"{self.output_dir_path}/"
            f"{f_name}"
        )
        
    @property
    def neg_lap_path(self) -> Path:
        if self.tag is None:
            save_name = '.neg_lap.pickle'
        else:
            save_name = '.'.join(['.neg_lap', self.tag, 'pickle'])
        f_name = re.sub(self.replace_term, save_name, self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )

    @property
    def eig_val_path(self) -> Path:
        if self.tag is None:
            save_name = '.eig_vals.pickle'
        else:
            save_name = '.'.join(['.eig_vals', self.tag, 'pickle'])
        f_name = re.sub(self.replace_term, save_name, self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )

    @property
    def eig_vec_path(self) -> Path:
        if self.tag is None:
            save_name = '.eig_vecs.txt.gz'
        else:
            save_name = '.'.join(['.eig_vecs', self.tag, 'txt.gz'])
        f_name = re.sub(self.replace_term, save_name, self.input_path.name)
        return Path(
            f"{self.output_dir_path}/" +
            f"{f_name}"
        )

    def run(self):
        print_arg("Number of simulations", self.num_eig)
        if self.eff_num_test:
            #self.eigen_decomposition()
            self.get_n_etests()
            self.update_env()
        else:
            self.eigen_decomposition()
        print_progress("Done")
    
    def get_n_etests(self):
        """Get the number of effective tests """
        print_progress(self.get_n_etests.__doc__)
        if os.path.isfile(self.eig_val_path):
            print_log(
                "NOTICE",
                "You already have eigen values for the selected categories.",
                False,
            )
        else:
            self.eigen_decomposition(save_vecs=False)
            
        with self.eig_val_path.open('rb') as f:
            eig_vals = pickle.load(f)
        
        e = 1e-12
        eig_vals = sorted(eig_vals, key=np.linalg.norm, reverse=True)
        num_eig_val = self.num_eig
        clean_eig_vals = np.array(eig_vals[:num_eig_val])
        clean_eig_vals = clean_eig_vals[clean_eig_vals >= e]
        clean_eig_val_total_sum = np.sum(clean_eig_vals)
        clean_eig_val_sum = 0
        eff_num_test = 0
        
        for i in range(len(clean_eig_vals)):
            clean_eig_val_sum += clean_eig_vals[i]

            if clean_eig_val_sum / clean_eig_val_total_sum >= 0.99:
                eff_num_test = i + 1
                break
        
        print_log("RESULT", f"The number of effective tests is {eff_num_test}.", False)
        
        self.eff_num_test_value = eff_num_test
        
    def eigen_decomposition(self, save_vecs: bool = True):
        print_progress(f"Calculate eigen values")
        if self.category_set_path:
            c1 = self.category_set["Category"].tolist()
        else:
            if self.input_format == 'corr':
                c1 = self.correlation_matrix.columns.tolist()
            elif self.input_format == 'inter':
                c1 = self.intersection_matrix.columns.tolist()

        print_progress(f"Categories with at least {self.count_thres} counts will be used")
        c2 = self.category_count[self.category_count['Raw_counts'] >= self.count_thres]['Category'].tolist()
        
        filtered_combs = [x for x in c1 if x in c2]
        
        print_progress(f"Use # of categories: {len(filtered_combs)}")

        if self.input_format == 'corr':
            intermediate_mat = self.correlation_matrix.loc[filtered_combs,filtered_combs]
        elif self.input_format == 'inter':
            print_progress("Generating a covariance matrix")
            intermediate_mat = self.intersection_matrix.loc[filtered_combs,filtered_combs]
            intermediate_mat = intermediate_mat.mul((self.binom_p)*(1-self.binom_p))

        if not os.path.isfile(self.neg_lap_path):
            print_progress("Generating the negative laplacian matrix")
            neg_lap = np.abs(intermediate_mat.values)
            degrees = np.sum(neg_lap, axis=0)
            for i in tqdm(range(neg_lap.shape[0])):
                neg_lap[i, :] = neg_lap[i, :] / np.sqrt(degrees)
                neg_lap[:, i] = neg_lap[:, i] / np.sqrt(degrees)
            print_progress("Writing the negative laplacian matrix to file")
            pickle.dump(neg_lap, open(self.neg_lap_path, 'wb'), protocol=5)
        else:
            with self.neg_lap_path.open('rb') as f:
                neg_lap = pickle.load(f)

        if not os.path.isfile(self.eig_val_path) or not os.path.isfile(self.eig_vec_path):
            print_progress("Calculating the eigenvalues of the negative laplacian matrix")
            eig_vals, eig_vecs = np.linalg.eig(neg_lap)
            print_progress("Writing the eigenvalues to file")
            pickle.dump(eig_vals, open(self.eig_val_path, 'wb'), protocol=5)
            if save_vecs:
                print_progress("Writing the eigenvectors to file")
                pd.DataFrame(eig_vecs.real, index=filtered_combs).to_csv(self.eig_vec_path, sep='\t', index=True, header=False)
                
                
    def update_env(self):
        self.set_env("N_EFFECTIVE_TEST", self.eff_num_test_value)
        self.save_env()
    
    


"""
Utils to execute system command line.
"""
import shutil
import subprocess
from pathlib import Path
from typing import Optional

import cwas.utils.log as log


class CmdExecutor:
    def __init__(self, bin: str, args: list = [], multi_input: Optional[str] = None, shell: bool = False):
        def resolve_bin(bin: str) -> Optional[str]:
            if Path(bin).resolve().is_file():
                return str(Path(bin).resolve())
            return shutil.which(bin)

        self._bin_path = resolve_bin(bin)
        if self._bin_path is None:
            raise FileNotFoundError(f"Failed to find the binary '{bin}'")
        self._args = list(args)
        self._multi_input = multi_input if multi_input is not None else None
        self._shell = shell

    @property
    def bin_path(self):
        return self._bin_path

    @property
    def cmd(self) -> list:
        if self._multi_input is None:
            cmd_args_list = [self._bin_path, *self._args]
        else:
            cmd_args_list = [self._multi_input, self._bin_path, *self._args]
        return cmd_args_list

    def execute(self) -> int:
        log.print_log("CMD", " ".join(self.cmd), True)
        output = subprocess.run(self.cmd, check=False)

        if output.returncode:
            log.print_warn(
                f"Command {self.cmd} has failed"
                f"with exit status {output.returncode}."
            )
        return output.returncode

    def execute_raising_err(self) -> int:
        if self._shell == True:
            log.print_log("CMD", " ".join(self.cmd), True)
            output = subprocess.run(" ".join(self.cmd), check=True, shell = True)            
        else:
            log.print_log("CMD", " ".join(self.cmd), True)
            output = subprocess.run(self.cmd, check=True, shell = False)
        return output.returncode

def compress_using_bgzip(
    input_file_path: Path, force_overwrite: bool = False
) -> Path:
    result = Path(str(input_file_path) + ".gz")
    if not force_overwrite and result.exists():
        log.print_warn(
            f'The compressed file "{result}" already exists. '
            f"Skip compressing."
        )
    else:
        args = [str(input_file_path)]
        if force_overwrite:
            args.append("-f")
        CmdExecutor("bgzip", args).execute_raising_err()

    return result


def index_using_tabix(
    compressed_file_path: Path, force_overwrite: bool = False
) -> Path:
    result = Path(str(compressed_file_path) + ".tbi")
    if not force_overwrite and result.exists():
        log.print_warn(
            f'The file index "{result}" already exists. Skip indexing.'
        )
    else:
        CmdExecutor("tabix", [str(compressed_file_path)]).execute_raising_err()

    return result
